"""Unit tests for the UserResource API."""

from __future__ import annotations

import logging

from django.contrib.auth.models import Permission, User
from djblets.avatars.services.base import AvatarService
from djblets.avatars.services.gravatar import GravatarService
from djblets.testing.decorators import add_fixtures
from djblets.webapi.errors import DOES_NOT_EXIST
from djblets.webapi.testing.decorators import webapi_test_template
from kgb import SpyAgency

from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.backends.registry import get_enabled_auth_backends
from reviewboard.avatars import avatar_services
from reviewboard.avatars.testcase import AvatarServicesTestMixin
from reviewboard.site.models import LocalSite
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (user_item_mimetype,
                                                user_list_mimetype)
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.urls import (get_user_item_url,
                                           get_user_list_url)


class NoProfileAuthBackend(BaseAuthBackend):
    backend_id = 'test-no-profile'
    supports_change_name = False
    supports_change_email = False


class BrokenUpdateProfileAuthBackend(BaseAuthBackend):
    backend_id = 'test-broken-profile'
    supports_change_name = True
    supports_change_email = True

    def update_name(self, user):
        raise Exception('oh no')

    def update_email(self, user):
        raise Exception('oh no')


class NoURLAvatarService(AvatarService):
    """An avatar services that returns no URLs."""

    avatar_service_id = 'no-urls'
    name = 'No URLs For You'

    def get_avatar_urls_uncached(self, user, size):
        """Return no URLs."""
        return {}


class SimpleRenderAvatarService(NoURLAvatarService):
    """An avatar services that has simple, testable rendered output."""

    avatar_service_id = 'simple-renders'
    name = 'Simple Renders'

    def render(self, request, user, size, **kwargs):
        return '<div class="avatar" data-size="%s">%s</div>' % (size,
                                                                user.username)


class ResourceListTests(AvatarServicesTestMixin, SpyAgency,
                        BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the UserResource list API tests."""

    fixtures = ['test_users']
    sample_api_url = 'users/'
    resource = resources.user

    test_http_methods = ('GET',)

    def setup_http_not_allowed_list_test(self, user):
        return get_user_list_url()

    def compare_item(self, item_rsp, obj):
        self.assertEqual(item_rsp['id'], obj.pk)
        self.assertEqual(item_rsp['username'], obj.username)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        if not populate_items:
            items = []
        elif with_local_site:
            local_site = self.get_local_site(name=local_site_name)
            items = list(local_site.users.all())
        else:
            items = list(User.objects.all())

        return (get_user_list_url(local_site_name),
                user_list_mimetype,
                items)

    @webapi_test_template
    def test_get_filter_inactive(self):
        """Testing the GET <URL> API filters out inactive users by default"""
        dopey = User.objects.get(username='dopey')
        dopey.is_active = False
        dopey.save()

        rsp = self.api_get(get_user_list_url(),
                           expected_mimetype=user_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        user_pks = [user['id'] for user in rsp['users']]
        returned_users = set(User.objects.filter(pk__in=user_pks))
        expected_users = set(User.objects.filter(is_active=True))
        self.assertEqual(returned_users, expected_users)

    @webapi_test_template
    def test_get_include_inactive(self):
        """Testing the GET <URL>/?include-inactive=1 API includes inactive
        users
        """
        dopey = User.objects.get(username='dopey')
        dopey.is_active = False
        dopey.save()

        rsp = self.api_get(get_user_list_url(), {'include-inactive': '1'},
                           expected_mimetype=user_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        user_pks = [user['id'] for user in rsp['users']]
        self.assertEqual(set(User.objects.filter(pk__in=user_pks)),
                         set(User.objects.all()))

    @webapi_test_template
    def test_get_include_inactive_true(self):
        """Testing the GET <URL>/?include-inactive=true API includes inactive
        users
        """
        dopey = User.objects.get(username='dopey')
        dopey.is_active = False
        dopey.save()

        rsp = self.api_get(get_user_list_url(), {'include-inactive': 'true'},
                           expected_mimetype=user_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        user_pks = [user['id'] for user in rsp['users']]
        self.assertEqual(set(User.objects.filter(pk__in=user_pks)),
                         set(User.objects.all()))

    def test_get_with_q(self):
        """Testing the GET users/?q= API"""
        rsp = self.api_get(get_user_list_url(), {'q': 'gru'},
                           expected_mimetype=user_list_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(len(rsp['users']), 1)  # grumpy

    @webapi_test_template
    def test_get_with_render_avatars_at(self):
        """Testing the GET <URL> API with ?render-avatars-at=..."""
        avatar_services.register(SimpleRenderAvatarService)
        avatar_services.enable_service(SimpleRenderAvatarService, save=False)
        avatar_services.set_default_service(SimpleRenderAvatarService)

        rsp = self.api_get(
            get_user_list_url(),
            {
                'render-avatars-at': '24,abc,48,,128',
            },
            expected_mimetype=user_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(len(rsp['users']), 4)
        self.assertEqual(
            rsp['users'][0]['avatar_html'],
            {
                '24': '<div class="avatar" data-size="24">admin</div>',
                '48': '<div class="avatar" data-size="48">admin</div>',
                '128': '<div class="avatar" data-size="128">admin</div>',
            })

        self.assertEqual(
            rsp['users'][1]['avatar_html'],
            {
                '24': '<div class="avatar" data-size="24">doc</div>',
                '48': '<div class="avatar" data-size="48">doc</div>',
                '128': '<div class="avatar" data-size="128">doc</div>',
            })

        self.assertEqual(
            rsp['users'][2]['avatar_html'],
            {
                '24': '<div class="avatar" data-size="24">dopey</div>',
                '48': '<div class="avatar" data-size="48">dopey</div>',
                '128': '<div class="avatar" data-size="128">dopey</div>',
            })

        self.assertEqual(
            rsp['users'][3]['avatar_html'],
            {
                '24': '<div class="avatar" data-size="24">grumpy</div>',
                '48': '<div class="avatar" data-size="48">grumpy</div>',
                '128': '<div class="avatar" data-size="128">grumpy</div>',
            })

    def test_populate_users_auth_backend(self):
        """Testing the GET users/?q= API with BaseAuthBackend.populate_users
        failure
        """
        class SandboxAuthBackend(BaseAuthBackend):
            backend_id = 'test-id'
            name = 'test'

            def populate_users(self, query, request, **kwargs):
                raise Exception

        backend = SandboxAuthBackend()

        self.spy_on(get_enabled_auth_backends, call_fake=lambda: [backend])
        self.spy_on(backend.populate_users)

        rsp = self.api_get(get_user_list_url(), {'q': 'gru'},
                           expected_mimetype=user_list_mimetype)
        self.assertEqual(rsp['stat'], 'ok')

        self.assertTrue(backend.populate_users.called)

    def test_build_search_users_query_auth_backend(self):
        """Testing the GET users/?q= API with
        BaseAuthBackend.build_search_users_query failure
        """
        class SandboxAuthBackend(BaseAuthBackend):
            backend_id = 'test-id'
            name = 'test'

            def build_search_users_query(self, query, request, **kwargs):
                raise Exception

        backend = SandboxAuthBackend()

        self.spy_on(get_enabled_auth_backends, call_fake=lambda: [backend])
        self.spy_on(backend.build_search_users_query)

        rsp = self.api_get(get_user_list_url(), {'q': 'gru'},
                           expected_mimetype=user_list_mimetype)
        self.assertEqual(rsp['stat'], 'ok')

        self.assertTrue(backend.build_search_users_query.called)

    #
    # HTTP POST tests
    #
    @webapi_test_template
    def test_post_anonymous(self):
        """Testing the POST <URL> API as an anonymous user"""
        self.client.logout()
        rsp = self.api_post(
            get_user_list_url(),
            {
                'username': 'username',
                'password': 'password',
                'email': 'email@example.com',
            },
            expected_status=401)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('err', rsp)
        self.assertIn('code', rsp['err'])
        self.assertEqual(rsp['err']['code'], 103)

    @webapi_test_template
    def test_post(self):
        """Testing the POST <URL> API as a regular user"""
        rsp = self.api_post(
            get_user_list_url(),
            {
                'username': 'username',
                'password': 'password',
                'email': 'email@example.com'
            },
            expected_status=403)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('err', rsp)
        self.assertIn('code', rsp['err'])
        self.assertEqual(rsp['err']['code'], 101)

    @webapi_test_template
    def test_post_superuser(self):
        """Testing the POST <URL> API as a superuser"""
        self.client.login(username='admin', password='admin')

        rsp = self.api_post(
            get_user_list_url(),
            {
                'username': 'username',
                'password': 'password',
                'email': 'email@example.com',
            },
            expected_mimetype=user_item_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.compare_item(rsp['user'], User.objects.get(username='username'))

    @webapi_test_template
    def test_post_auth_add_user_perm(self):
        """Testing the POST <URL> API as a user with the auth.add_user
        permission
        """
        self.user.user_permissions.add(
            Permission.objects.get(content_type__app_label='auth',
                                   codename='add_user'))

        rsp = self.api_post(
            get_user_list_url(),
            {
                'username': 'username',
                'password': 'password',
                'email': 'email@example.com',
            },
            expected_mimetype=user_item_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.compare_item(rsp['user'], User.objects.get(username='username'))

    @webapi_test_template
    def test_post_local_site(self):
        """Testing the POST <URL> API with a local site"""
        local_site = LocalSite.objects.create(name='test', public=True)

        self.client.login(username='admin', password='admin')
        rsp = self.api_post(
            get_user_list_url(local_site.name),
            {
                'username': 'username',
                'password': 'password',
                'email': 'email@example.com'
            },
            expected_status=403)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('err', rsp)
        self.assertIn('code', rsp['err'])
        self.assertEqual(rsp['err']['code'], 101)

    @webapi_test_template
    def test_post_duplicate_username(self):
        """Testing the POST <URL> API for a username that already exists"""
        self.client.login(username='admin', password='admin')
        rsp = self.api_post(
            get_user_list_url(),
            {
                'username': 'doc',
                'password': 'password',
                'email': 'doc@example.com'
            },
            expected_status=400)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('fields', rsp)
        self.assertIn('username', rsp['fields'])

    @webapi_test_template
    def test_post_invalid_email(self):
        """Testing the POST <URL> API for an invalid e-mail address"""
        self.client.login(username='admin', password='admin')
        rsp = self.api_post(
            get_user_list_url(),
            {
                'username': 'username',
                'password': 'password',
                'email': 'invalid e-mail',
            },
            expected_status=400)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('fields', rsp)
        self.assertEqual(rsp['fields'], {
            'email': ['Enter a valid email address.'],
        })

    @webapi_test_template
    def test_post_with_render_avatars_at(self):
        """Testing the POST <URL> API with render_avatars_at=..."""
        self.client.login(username='admin', password='admin')

        avatar_services.register(SimpleRenderAvatarService)
        avatar_services.enable_service(SimpleRenderAvatarService, save=False)
        avatar_services.set_default_service(SimpleRenderAvatarService)

        rsp = self.api_post(
            get_user_list_url(),
            {
                'username': 'myuser',
                'password': 'mypass',
                'email': 'myuser@example.com',
                'render_avatars_at': '24,abc,48,,128',
            },
            expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(
            rsp['user']['avatar_html'],
            {
                '24': '<div class="avatar" data-size="24">myuser</div>',
                '48': '<div class="avatar" data-size="48">myuser</div>',
                '128': '<div class="avatar" data-size="128">myuser</div>',
            })

    @webapi_test_template
    def test_post_without_password(self) -> None:
        """Testing the POST <URL> API with no password"""
        self.client.login(username='admin', password='admin')
        rsp = self.api_post(
            get_user_list_url(),
            {
                'username': 'myuser',
                'email': 'myuser@example.com',
            },
            expected_mimetype=user_item_mimetype,
        )

        assert rsp is not None
        self.assertEqual(rsp['stat'], 'ok')

        user = User.objects.get(username='myuser')
        self.assertFalse(user.has_usable_password())


class ResourceItemTests(AvatarServicesTestMixin, SpyAgency,
                        BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the UserResource item API tests."""
    fixtures = ['test_users']
    sample_api_url = 'users/<username>/'
    resource = resources.user
    test_http_methods = ('GET',)

    def setUp(self):
        super(ResourceItemTests, self).setUp()

        avatar_services.enable_service(GravatarService, save=False)

    def setup_http_not_allowed_item_test(self, user):
        return get_user_item_url(user.username)

    def compare_item(self, item_rsp, user):
        self.assertEqual(item_rsp['id'], user.pk)
        self.assertEqual(item_rsp['username'], user.username)
        self.assertEqual(item_rsp['first_name'], user.first_name)
        self.assertEqual(item_rsp['last_name'], user.last_name)
        self.assertEqual(item_rsp['email'], user.email)

        # There's no simple way to test the specific URLs that are returned,
        # but we can at least make sure everything we expect to be present is
        # present.
        self.assertIn('avatar_url', item_rsp)
        self.assertIn('1x', item_rsp['avatar_urls'])
        self.assertIn('2x', item_rsp['avatar_urls'])

        # By default, avatars are not rendered.
        self.assertIsNone(item_rsp['avatar_html'])

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        return (get_user_item_url(user.username, local_site_name),
                user_item_mimetype,
                user)

    def test_get_not_modified(self):
        """Testing the GET users/<username>/ API with Not Modified response"""
        self._testHttpCaching(get_user_item_url('doc'),
                              check_etags=True)

    def test_get_with_site_and_profile_private(self):
        """Testing the GET users/<username>/ API with a local site and private
        profile
        """
        username = 'admin'
        user = User.objects.get(username=username)

        site = LocalSite.objects.create(name=self.local_site_name)
        site.users.add(user, self.user)

        profile = user.get_profile()
        profile.is_private = True
        profile.save(update_fields=('is_private',))

        rsp = self.api_get(get_user_item_url(username, self.local_site_name),
                           expected_mimetype=user_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['user']['username'], user.username)
        self.assertNotIn('first_name', rsp['user'])
        self.assertNotIn('last_name', rsp['user'])
        self.assertNotIn('email', rsp['user'])

    @add_fixtures(['test_site'])
    def test_get_with_site_and_profile_private_as_site_admin(self):
        """Testing the GET users/<username>/ API with a local site and private
        profile as a LocalSite admin
        """
        self._login_user(local_site=True)

        username = 'admin'
        user = User.objects.get(username=username)

        profile = user.get_profile()
        profile.is_private = True
        profile.save(update_fields=('is_private',))

        rsp = self.api_get(get_user_item_url(username, self.local_site_name),
                           expected_mimetype=user_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['user']
        self.assertEqual(item_rsp['username'], user.username)
        self.assertEqual(item_rsp['first_name'], user.first_name)
        self.assertEqual(item_rsp['last_name'], user.last_name)
        self.assertEqual(item_rsp['email'], user.email)

    @add_fixtures(['test_site'])
    def test_get_missing_user_with_site(self):
        """Testing the GET users/<username>/ API with a local site"""
        self._login_user(local_site=True)
        self.api_get(get_user_item_url('dopey', self.local_site_name),
                     expected_status=404)

    @webapi_test_template
    def test_get_with_profile_private_and_only_fields(self):
        """Testing the GET <URL> API with a private profile and ?only-fields=
        """
        username = 'dopey'
        user = User.objects.get(username=username)

        profile = user.get_profile()
        profile.is_private = True
        profile.save(update_fields=('is_private',))

        rsp = self.api_get(
            '%s?only-fields=username' % get_user_item_url(username),
            expected_mimetype=user_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['user']['username'], user.username)
        self.assertNotIn('first_name', rsp['user'])
        self.assertNotIn('last_name', rsp['user'])
        self.assertNotIn('email', rsp['user'])

    @webapi_test_template
    def test_get_inactive_user(self):
        """Testing the GET <URL> API for an inactive user"""
        dopey = User.objects.get(username='dopey')
        dopey.is_active = False
        dopey.save()

        rsp = self.api_get(get_user_item_url('dopey'),
                           expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['user']['is_active'], False)

    @webapi_test_template
    def test_get_avatar_service_no_urls(self):
        """Testing the GET <URL> API when the avatar service returns no URLs
        """
        avatar_services.register(NoURLAvatarService)
        avatar_services.enable_service(NoURLAvatarService, save=False)

        dopey = User.objects.get(username='dopey')
        settings_mgr = avatar_services.settings_manager_class(dopey)
        settings_mgr.avatar_service_id = NoURLAvatarService.avatar_service_id
        settings_mgr.save()

        rsp = self.api_get(get_user_item_url('dopey'),
                           expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        user_rsp = rsp['user']
        self.assertIn('avatar_url', user_rsp)
        self.assertIsNone(user_rsp['avatar_url'])
        self.assertIn('avatar_urls', user_rsp)
        self.assertEqual(user_rsp['avatar_urls'], {})

    @webapi_test_template
    def test_get_with_render_avatars_at(self):
        """Testing the GET <URL> API with ?render-avatars-at=..."""
        avatar_services.register(SimpleRenderAvatarService)
        avatar_services.enable_service(SimpleRenderAvatarService, save=False)
        avatar_services.set_default_service(SimpleRenderAvatarService)

        rsp = self.api_get(
            get_user_item_url('dopey'),
            {
                'render-avatars-at': '24,abc,48,,128',
            },
            expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(
            rsp['user']['avatar_html'],
            {
                '24': '<div class="avatar" data-size="24">dopey</div>',
                '48': '<div class="avatar" data-size="48">dopey</div>',
                '128': '<div class="avatar" data-size="128">dopey</div>',
            })

    #
    # HTTP PUT tests
    #

    @webapi_test_template
    def test_put_with_user_not_found(self):
        """Testing the PUT <URL> API with username not found"""
        self._login_user(admin=True)

        rsp = self.api_put(
            get_user_item_url('bad-username'),
            {
                'first_name': 'new-first-name',
            },
            expected_status=404)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DOES_NOT_EXIST.code)

    @webapi_test_template
    def test_put_profile_fields_as_same_user(self):
        """Testing the PUT <URL> API with profile fields as user being
        modified
        """
        auth_backend = get_enabled_auth_backends()[0]
        self.spy_on(auth_backend.update_name)
        self.spy_on(auth_backend.update_email)

        user = self.create_user(username='test-user',
                                password='test-user')
        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'first_name': 'new-first-name',
                'last_name': 'new-last-name',
                'email': 'new-email@example.com',
            },
            expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        user_rsp = rsp['user']
        self.assertEqual(user_rsp['first_name'], 'new-first-name')
        self.assertEqual(user_rsp['last_name'], 'new-last-name')
        self.assertEqual(user_rsp['email'], 'new-email@example.com')

        user = User.objects.get(pk=user.pk)
        self.assertEqual(user.first_name, 'new-first-name')
        self.assertEqual(user.last_name, 'new-last-name')
        self.assertEqual(user.email, 'new-email@example.com')

        self.assertTrue(auth_backend.update_name.called_with(user))
        self.assertTrue(auth_backend.update_email.called_with(user))

    @webapi_test_template
    def test_put_profile_fields_as_other_user(self):
        """Testing the PUT <URL> API with profile fields as other user"""
        user = self.create_user(username='test-user',
                                password='test-user')

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'first_name': 'new-first-name',
                'last_name': 'new-last-name',
                'email': 'new-email@example.com',
            },
            expected_status=403)

        self.assertEqual(rsp['stat'], 'fail')

    @webapi_test_template
    def test_put_profile_fields_as_superuser(self):
        """Testing the PUT <URL> API with profile fields as superuser"""
        auth_backend = get_enabled_auth_backends()[0]
        self.spy_on(auth_backend.update_name)
        self.spy_on(auth_backend.update_email)

        user = self.create_user(username='test-user',
                                password='test-user')
        self._login_user(admin=True)

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'first_name': 'new-first-name',
                'last_name': 'new-last-name',
                'email': 'new-email@example.com',
            },
            expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        user_rsp = rsp['user']
        self.assertEqual(user_rsp['first_name'], 'new-first-name')
        self.assertEqual(user_rsp['last_name'], 'new-last-name')
        self.assertEqual(user_rsp['email'], 'new-email@example.com')

        user = User.objects.get(pk=user.pk)
        self.assertEqual(user.first_name, 'new-first-name')
        self.assertEqual(user.last_name, 'new-last-name')
        self.assertEqual(user.email, 'new-email@example.com')

        self.assertTrue(auth_backend.update_name.called_with(user))
        self.assertTrue(auth_backend.update_email.called_with(user))

    @webapi_test_template
    def test_put_profile_fields_as_user_with_perm(self):
        """Testing the PUT <URL> API with profile fields as user with
        auth.change_user permission
        """
        auth_backend = get_enabled_auth_backends()[0]
        self.spy_on(auth_backend.update_name)
        self.spy_on(auth_backend.update_email)

        user = self.create_user(username='test-user',
                                password='test-user')

        self.create_user(username='login-user',
                         password='login-user',
                         perms=[('auth', 'change_user')])
        self.assertTrue(self.client.login(username='login-user',
                                          password='login-user'))

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'first_name': 'new-first-name',
                'last_name': 'new-last-name',
                'email': 'new-email@example.com',
            },
            expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        user_rsp = rsp['user']
        self.assertEqual(user_rsp['first_name'], 'new-first-name')
        self.assertEqual(user_rsp['last_name'], 'new-last-name')
        self.assertEqual(user_rsp['email'], 'new-email@example.com')

        user = User.objects.get(pk=user.pk)
        self.assertEqual(user.first_name, 'new-first-name')
        self.assertEqual(user.last_name, 'new-last-name')
        self.assertEqual(user.email, 'new-email@example.com')

        self.assertTrue(auth_backend.update_name.called_with(user))
        self.assertTrue(auth_backend.update_email.called_with(user))

    @webapi_test_template
    def test_put_with_invalid_email_field(self):
        """Testing the PUT <URL> API with invalid e-mail field"""
        user = self.create_user(username='test-user',
                                password='test-user')
        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'email': 'bad-email',
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('fields', rsp)
        self.assertEqual(rsp['fields'], {
            'email': ['Enter a valid email address.'],
        })

    @webapi_test_template
    def test_put_with_full_name_and_no_backend_support(self):
        """Testing the PUT <URL> API with setting first_name or last_name
        with auth_backend.supports_change_name == False
        """
        self.spy_on(get_enabled_auth_backends,
                    call_fake=lambda: [NoProfileAuthBackend()])

        user = self.create_user(username='test-user',
                                password='test-user')
        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'first_name': 'new_first_name',
                'last_name': 'new_last_name',
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('fields', rsp)
        self.assertEqual(rsp['fields'], {
            'first_name': ['The configured auth backend does not allow names '
                           'to be changed.'],
            'last_name': ['The configured auth backend does not allow names '
                          'to be changed.'],
        })

    @webapi_test_template
    def test_put_with_email_and_no_backend_support(self):
        """Testing the PUT <URL> API with setting email with
        auth_backend.supports_change_email == False
        """
        self.spy_on(get_enabled_auth_backends,
                    call_fake=lambda: [NoProfileAuthBackend()])

        user = self.create_user(username='test-user',
                                password='test-user')
        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'email': 'new@example.com',
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('fields', rsp)
        self.assertEqual(rsp['fields'], {
            'email': ['The configured auth backend does not allow e-mail '
                      'addresses to be changed.'],
        })

    @webapi_test_template
    def test_put_with_update_namel_failure(self):
        """Testing the PUT <URL> API with auth_backend.update_name() failure"""
        auth_backend = BrokenUpdateProfileAuthBackend()
        logger = logging.getLogger('reviewboard.webapi.resources.user')

        self.spy_on(get_enabled_auth_backends,
                    call_fake=lambda: [auth_backend])
        self.spy_on(logger.exception)

        user = self.create_user(username='test-user',
                                password='test-user')
        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'first_name': 'new-first-name',
                'last_name': 'new-last-name',
            },
            expected_mimetype=user_item_mimetype)

        user_rsp = rsp['user']
        self.assertEqual(user_rsp['first_name'], 'new-first-name')
        self.assertEqual(user_rsp['last_name'], 'new-last-name')

        user = User.objects.get(pk=user.pk)
        self.assertEqual(user.first_name, 'new-first-name')
        self.assertEqual(user.last_name, 'new-last-name')

        self.assertTrue(logger.exception.called_with(
            'Error when calling update_name for auth backend %r for user '
            'ID %s: %s',
            auth_backend, user.pk))

    @webapi_test_template
    def test_put_with_update_email_failure(self):
        """Testing the PUT <URL> API with auth_backend.update_email() failure
        """
        auth_backend = BrokenUpdateProfileAuthBackend()
        logger = logging.getLogger('reviewboard.webapi.resources.user')

        self.spy_on(get_enabled_auth_backends,
                    call_fake=lambda: [auth_backend])
        self.spy_on(logger.exception)

        user = self.create_user(username='test-user',
                                password='test-user')
        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'email': 'new@example.com',
            },
            expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['user']['email'], 'new@example.com')

        user = User.objects.get(pk=user.pk)
        self.assertEqual(user.email, 'new@example.com')

        self.assertTrue(logger.exception.called_with(
            'Error when calling update_email for auth backend %r for user '
            'ID %s: %s',
            auth_backend, user.pk))

    @webapi_test_template
    def test_put_is_active_as_same_user(self):
        """Testing the PUT <URL> API with is_active field as user being
        modified
        """
        user = self.create_user(username='test-user',
                                password='test-user')
        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'is_active': 'false',
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('fields', rsp)
        self.assertEqual(rsp['fields'], {
            'is_active': [
                'This field can only be set by administrators and users '
                'with the auth.change_user permission.'
            ],
        })

        user = User.objects.get(pk=user.pk)
        self.assertTrue(user.is_active)

    @webapi_test_template
    def test_put_is_active_as_superuser(self):
        """Testing the PUT <URL> API with is_active field as superuser"""
        user = self.create_user(username='test-user',
                                password='test-user')
        self._login_user(admin=True)

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'is_active': 'false',
            },
            expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        user_rsp = rsp['user']
        self.assertFalse(user_rsp['is_active'])

        user = User.objects.get(pk=user.pk)
        self.assertFalse(user.is_active)

    @webapi_test_template
    def test_put_is_active_as_user_with_perm(self):
        """Testing the PUT <URL> API with profile fields as user with
        auth.change_user permission
        """
        user = self.create_user(username='test-user',
                                password='test-user')

        self.create_user(username='login-user',
                         password='login-user',
                         perms=[('auth', 'change_user')])
        self.assertTrue(self.client.login(username='login-user',
                                          password='login-user'))

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'is_active': 'false',
            },
            expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        user_rsp = rsp['user']
        self.assertFalse(user_rsp['is_active'])

        user = User.objects.get(pk=user.pk)
        self.assertFalse(user.is_active)

    @webapi_test_template
    def test_put_with_render_avatars_at(self):
        """Testing the PUT <URL> API with render_avatars_at=..."""
        avatar_services.register(SimpleRenderAvatarService)
        avatar_services.enable_service(SimpleRenderAvatarService, save=False)
        avatar_services.set_default_service(SimpleRenderAvatarService)

        user = self.create_user(username='test-user',
                                password='test-user')
        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        rsp = self.api_put(
            get_user_item_url(user.username),
            {
                'first_name': 'new-name',
                'render_avatars_at': '24,abc,48,,128',
            },
            expected_mimetype=user_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(
            rsp['user']['avatar_html'],
            {
                '24': '<div class="avatar" data-size="24">test-user</div>',
                '48': '<div class="avatar" data-size="48">test-user</div>',
                '128': '<div class="avatar" data-size="128">test-user</div>',
            })



"""A Review UI for image files."""

from __future__ import annotations

from typing import Optional, Sequence, TYPE_CHECKING
from urllib.parse import urlparse

from django.utils.html import escape
from djblets.util.templatetags.djblets_images import crop_image

from reviewboard.admin.server import build_server_url
from reviewboard.attachments.models import FileAttachment
from reviewboard.reviews.models import FileAttachmentComment
from reviewboard.reviews.ui.base import (ReviewUI,
                                         SerializedComment,
                                         SerializedCommentBlocks)


if TYPE_CHECKING:
    from typelets.json import JSONDict


class SerializedRegionComment(SerializedComment):
    """Serialized data for an image comment.

    This must be kept in sync with the definitions in
    :file:`reviewboard/static/rb/js/reviews/models/commentData.ts`.

    Version Added:
        7.0
    """

    #: The X position of the comment block, in pixels.
    x: int

    #: The Y position of the comment block, in pixels.
    y: int

    #: The width of the comment block, in pixels.
    width: int

    #: The height of the comment block, in pixels.
    height: int


class ImageReviewUI(ReviewUI[
    FileAttachment,
    FileAttachmentComment,
    SerializedRegionComment
]):
    """A Review UI for image files."""

    name = 'Image'
    supported_mimetypes = ['image/*']

    allow_inline = True
    supports_diffing = True
    supports_file_attachments = True

    js_model_class: str = 'RB.ImageReviewable'
    js_view_class: str = 'RB.ImageReviewableView'

    def get_page_cover_image_url(self):
        """Return the URL to an image used to depict this on other sites.

        The returned image URL will be used for services like Facebook, Slack,
        Twitter, etc. when linking to this file attachment.

        Returns:
            str:
            The absolute URL to an image used to depict this file attachment.
        """
        return self.obj.get_absolute_url()

    def get_js_model_data(self) -> JSONDict:
        """Return data to pass to the JavaScript Model during instantiation.

        This data will be passed as attributes to the reviewable model when
        constructed.

        Returns:
            dict:
            The attributes to pass to the model.
        """
        data = super(ImageReviewUI, self).get_js_model_data()
        data['imageURL'] = self.obj.get_absolute_url()

        if self.diff_against_obj:
            assert isinstance(self.diff_against_obj, FileAttachment)
            data['diffAgainstImageURL'] = \
                self.diff_against_obj.get_absolute_url()

        return data

    def serialize_comments(
        self,
        comments: Sequence[FileAttachmentComment],
    ) -> SerializedCommentBlocks[SerializedRegionComment]:
        """Serialize the comments for the file attachment.

        Args:
            comments (list of
                      reviewboard.reviews.models.FileAttachmentComment):
                The list of objects to serialize. This will be the result of
                :py:meth:`get_comments`.

        Returns:
            SerializedCommentBlocks:
            The serialized comment data.
        """
        result: SerializedCommentBlocks[SerializedRegionComment] = {}

        for comment in self.flat_serialized_comments(comments):
            try:
                position = (
                    f'{comment["x"]}x{comment["y"]}+'
                    f'{comment["width"]}+{comment["height"]}')
            except KeyError:
                # It's possible this comment was made before the review UI
                # was provided, meaning it has no data. If this is the case,
                # ignore this particular comment, since it doesn't have a
                # region.
                continue

            result.setdefault(position, []).append(comment)

        return result

    def get_comment_thumbnail(
        self,
        comment: FileAttachmentComment,
    ) -> Optional[str]:
        """Generate and return a thumbnail representing this comment.

        This will find the appropriate lines the comment applies to and
        return it as HTML suited for rendering in reviews.

        Args:
            comment (reviewboard.reviews.models.FileAttachmentComment):
                The comment to render the thumbnail for.

        Returns:
            str:
            The rendered comment thumbnail.
        """
        try:
            x = int(comment.extra_data['x'])
            y = int(comment.extra_data['y'])
            width = int(comment.extra_data['width'])
            height = int(comment.extra_data['height'])
        except (KeyError, ValueError):
            # This may be a comment from before we had review UIs. Or,
            # corrupted data. Either way, don't display anything.
            return None

        image_url = crop_image(comment.file_attachment.file,
                               x, y, width, height)

        if not urlparse(image_url).netloc:
            image_url = build_server_url(image_url)

        image_html = (
            '<img class="modified-image" src="%s" width="%s" height="%s" '
            'alt="%s" />'
            % (image_url, width, height, escape(comment.text)))

        if comment.diff_against_file_attachment_id:
            diff_against_image_url = crop_image(
                comment.diff_against_file_attachment.file,
                x, y, width, height)

            if not urlparse(diff_against_image_url).netloc:
                diff_against_image_url = build_server_url(
                    diff_against_image_url)

            diff_against_image_html = (
                '<img class="orig-image" src="%s" width="%s" '
                'height="%s" alt="%s" />'
                % (diff_against_image_url, width, height,
                   escape(comment.text)))

            return ('<div class="image-review-ui-diff-thumbnail">%s%s</div>'
                    % (diff_against_image_html, image_html))
        else:
            return image_html



"""Extension hook for registering a license provider.

Version Added:
    8.0
"""

from __future__ import annotations

from djblets.extensions.hooks import (BaseRegistryHook,
                                      ExtensionHookPoint)

from reviewboard.licensing.provider import BaseLicenseProvider
from reviewboard.licensing.registry import license_provider_registry


class LicenseProviderHook(BaseRegistryHook[BaseLicenseProvider],
                          metaclass=ExtensionHookPoint):
    """Extension hook for registering a license provider.

    Version Added:
        8.0
    """

    registry = license_provider_registry



from django.dispatch import Signal


#: Emitted before a file existence check occurs.
#:
#: Args:
#:     path (unicode):
#:         The path of the file being checked.
#:
#:     revision (unicode):
#:         The revision of the file being checked.
#:
#:     request (django.http.HttpRequest):
#:         The current request.
checking_file_exists = Signal()


#: Emitted after a file existence check occurs.
#:
#: Args:
#:     path (unicode):
#:         The path of the file being checked.
#:
#:     revision (unicode):
#:         The revision of the file being checked.
#:
#:     request (django.http.HttpRequest):
#:         The current request.
checked_file_exists = Signal()


#: Emitted before a file is fetched.
#:
#: Args:
#:     path (unicode):
#:         The path of the file being checked.
#:
#:     revision (unicode):
#:         The revision of the file being checked.
#:
#:     request (django.http.HttpRequest):
#:         The current request.
fetching_file = Signal()


#: Emitted after a file is fetched.
#:
#: Args:
#:     path (unicode):
#:         The path of the file being checked.
#:
#:     revision (unicode):
#:         The revision of the file being checked.
#:
#:     request (django.http.HttpRequest):
#:         The current request.
fetched_file = Signal()



"""Hosting service for Bitbucket."""

from __future__ import annotations

import json
import logging
from collections import defaultdict
from typing import TYPE_CHECKING
from urllib.error import HTTPError
from urllib.parse import quote

from django import forms
from django.core.cache import cache
from django.http import (HttpResponse,
                         HttpResponseBadRequest,
                         HttpResponseForbidden)
from django.template.loader import render_to_string
from django.urls import path
from django.utils.translation import gettext_lazy as _, gettext
from django.views.decorators.http import require_POST
from housekeeping import deprecate_non_keyword_only_args

from reviewboard.admin.server import build_server_url, get_server_url
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.hostingsvcs.base.forms import (
    BaseHostingServiceAuthForm,
    BaseHostingServiceRepositoryForm,
)
from reviewboard.hostingsvcs.base.client import HostingServiceClient
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            HostingServiceAPIError,
                                            HostingServiceError,
                                            InvalidPlanError,
                                            RepositoryError)
from reviewboard.hostingsvcs.hook_utils import (close_all_review_requests,
                                                get_repository_for_hook,
                                                get_review_request_id)
from reviewboard.hostingsvcs.utils.paginator import APIPaginator
from reviewboard.scmtools.core import Branch, Commit
from reviewboard.scmtools.crypto_utils import encrypt_password
from reviewboard.scmtools.errors import (FileNotFoundError,
                                         RepositoryNotFoundError)
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


EMAIL_SETTINGS_URL = 'https://id.atlassian.com/manage-profile/email'
TOKEN_SETTINGS_URL = \
    'https://id.atlassian.com/manage-profile/security/api-tokens'


class BitbucketAuthForm(BaseHostingServiceAuthForm):
    """Authentication form for linking a Bitbucket account."""

    class Meta:
        """Metadata for the form."""

        help_texts = {
            'hosting_account_username': _(
                'Your Bitbucket account e-mail address. This can be found in '
                'your <a href="{email_settings_url}" target="_blank">'
                'Atlassian Account Settings</a>.'
            ).format(email_settings_url=EMAIL_SETTINGS_URL),
            'hosting_account_password': _(
                'A new <a href="{token_settings_url}" target="_blank">'
                'Atlassian API Token</a>. To create this, select "Create API '
                'token with scopes", set a long expiration time on the token, '
                'choose Bitbucket as the product, and then add scopes '
                '<code>read:user:bitbucket</code> and '
                '<code>read:repository:bitbucket</code>.'
            ).format(token_settings_url=TOKEN_SETTINGS_URL),
        }


class BitbucketPersonalForm(BaseHostingServiceRepositoryForm):
    bitbucket_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the repository. This is the '
                    '&lt;repo_name&gt; in https://bitbucket.org/'
                    '&lt;username&gt;/&lt;repo_name&gt;/'))


class BitbucketOtherUserForm(BaseHostingServiceRepositoryForm):
    bitbucket_other_user_username = forms.CharField(
        label=_('Username'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The username of the user who owns the repository. This '
                    'is the &lt;username&gt; in '
                    'https://bitbucket.org/&lt;username&gt;/'
                    '&lt;repo_name&gt;/'))

    bitbucket_other_user_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the repository. This is the '
                    '&lt;repo_name&gt; in '
                    'https://bitbucket.org/&lt;username&gt;/'
                    '&lt;repo_name&gt;/'))


class BitbucketTeamForm(BaseHostingServiceRepositoryForm):
    bitbucket_team_name = forms.CharField(
        label=_('Team name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the team. This is the &lt;team_name&gt; in '
                    'https://bitbucket.org/&lt;team_name&gt;/'
                    '&lt;repo_name&gt;/'))

    bitbucket_team_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the repository. This is the '
                    '&lt;repo_name&gt; in '
                    'https://bitbucket.org/&lt;team_name&gt;/'
                    '&lt;repo_name&gt;/'))


class BitbucketHookViews(object):
    """Container class for hook views."""

    @staticmethod
    @require_POST
    def post_receive_hook_close_submitted(
        request: HttpRequest,
        local_site_name: (str | None) = None,
        repository_id: (int | None) = None,
        hosting_service_id: (str | None) = None,
        hooks_uuid: (str | None) = None,
    ) -> HttpResponse:
        """Close review requests as submitted automatically after a push.

        Args:
            request (django.http.HttpRequest):
                The request from the Bitbucket webhook.

            local_site_name (str, optional):
                The local site name, if available.

            repository_id (int, optional):
                The pk of the repository, if available.

            hosting_service_id (str, optional):
                The name of the hosting service.

            hooks_uuid (str, optional):
                The UUID of the configured webhook.

        Returns:
            django.http.HttpResponse:
            A response for the request.
        """
        assert repository_id is not None
        assert hosting_service_id is not None

        repository = get_repository_for_hook(
            repository_id=repository_id,
            hosting_service_id=hosting_service_id,
            local_site_name=local_site_name,
            hooks_uuid=hooks_uuid)

        try:
            payload = json.loads(request.body.decode('utf-8'))
        except ValueError as e:
            logger.error('The payload is not in JSON format: %s', e,
                         extra={'request': request})
            return HttpResponseBadRequest('Invalid payload format')

        server_url = get_server_url(request=request)

        try:
            review_request_id_to_commits = \
                BitbucketHookViews._get_review_request_id_to_commits_map(
                    payload=payload,
                    server_url=server_url,
                    repository=repository)
        except AuthorizationError:
            return HttpResponseForbidden(
                'Incorrect username or password configured for this '
                'repository on Review Board.')

        if review_request_id_to_commits:
            close_all_review_requests(
                review_request_id_to_commits=review_request_id_to_commits,
                local_site_name=local_site_name,
                repository=repository,
                hosting_service_id=hosting_service_id)

        return HttpResponse()

    @classmethod
    def _get_review_request_id_to_commits_map(cls, payload, server_url,
                                              repository):
        """Return a mapping of review request ID to a list of commits.

        If a commit's commit message does not contain a review request ID, we
        append the commit to the key None.

        Args:
            payload (dict):
                The decoded webhook payload.

            server_url (unicode):
                The URL of the Review Board server.

            repository (reviewboard.scmtools.models.Repository):
                The repository object.

        Returns:
            dict:
            A mapping from review request ID to a list of matching commits from
            the payload.
        """
        results = defaultdict(list)

        try:
            changes = payload['push']['changes']
        except KeyError:
            return results

        seen_commits_urls = set()

        for change in changes:
            change_new = change.get('new') or {}

            if (change_new and
                change_new['type'] not in ('branch', 'named_branch',
                                           'bookmark')):
                continue

            # These should always be here, but we want to be defensive.
            truncated = change.get('truncated', False)
            commits = change.get('commits') or []
            target_name = change_new.get('name')

            if not target_name or not commits:
                continue

            if truncated:
                try:
                    commits_url = change['links']['commits']['href']
                except KeyError:
                    commits_url = None

                if commits_url is not None:
                    commits = cls._iter_commits(
                        repository.hosting_service,
                        commits_url,
                        seen_commits_urls=seen_commits_urls)

            for commit in commits:
                commit_hash = commit.get('hash')
                commit_message = commit.get('message')

                review_request_id = get_review_request_id(
                    commit_message=commit_message,
                    server_url=server_url,
                    commit_id=commit_hash,
                    repository=repository)

                if review_request_id is not None:
                    results[review_request_id].append(
                        '%s (%s)' % (target_name, commit_hash[:7]))

        return results

    @classmethod
    def _iter_commits(cls, hosting_service, commits_url, seen_commits_urls,
                      max_pages=5):
        """Iterate through all pages of commits for a URL.

        This will go through each page of commits corresponding to a Push
        event, yielding each commit for further processing.

        Args:
            hosting_service (Bitbucket):
                The hosting service instance.

            commits_url (unicode):
                The beginning URL to page through.

            seen_commits_urls (set):
                The URLs that have already been seen. If a URL from this set
                is encountered, pagination will stop.

            max_pages (int, optional):
                The maximum number of pages to iterate through.

        Yields:
            dict:
            A payload for an individual commit.
        """
        if commits_url in seen_commits_urls:
            return

        paginator = BitbucketAPIPaginator(client=hosting_service.client,
                                          url=commits_url,
                                          per_page=100)

        for page_data in paginator.iter_pages(max_pages=max_pages):
            seen_commits_urls.add(paginator.url)

            for commit_rsp in page_data:
                yield commit_rsp

            if paginator.next_url in seen_commits_urls:
                break


class BitbucketAPIPaginator(APIPaginator):
    """Paginator for multi-page API responses on Bitbucket.

    This is returned by some :py:class:`BitbucketClient` functions in order
    to handle iteration over pages of results.
    """

    start_query_param = 'page'
    per_page_query_param = 'pagelen'

    def fetch_url(self, url):
        """Fetch the page data for a URL.

        Args:
            url (unicode):
                The URL to fetch.

        Returns:
            dict:
            Information on the page of results.
        """
        response = self.client.http_get(url, **self.request_kwargs)
        rsp = response.json

        return {
            'data': rsp.get('values'),
            'headers': response.headers,
            'total_count': rsp.get('size'),
            'prev_url': rsp.get('previous'),
            'next_url': rsp.get('next'),
        }


class BitbucketClient(HostingServiceClient):
    """Client interface to the Bitbucket Cloud API."""

    def __init__(self, *args, **kwargs):
        """Initialize the client.

        Args:
            *args (tuple):
                Positional arguments for the parent class.

            **kwargs (dict):
                Keyword arguments for the parent class.
        """
        super(BitbucketClient, self).__init__(*args, **kwargs)

        self.api_url = 'https://api.bitbucket.org/2.0'

    def api_get_user_session(self, **kwargs):
        """Return information on the user's session.

        Args:
            **kwargs (dict):
                Additional keyword arguments to pass in the request.

        Returns:
            dict:
            The user session data.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the file contents. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = '%s/user' % self.api_url

        return self.http_get(url, **kwargs).json

    def api_get_repository(self, repo_owner, repo_name, **kwargs):
        """Return information on a repository.

        Args:
            repo_owner (unicode):
                The owner of the repository.

            repo_name (unicode):
                The name of the repository.

            **kwargs (dict):
                Additional keyword arguments to pass in the request.

        Returns:
            dict:
            The repository information.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the file contents. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = self._get_repositories_api_url(repo_owner, repo_name)

        return self.http_get(url, **kwargs).json

    def api_get_file_contents(self, repo_owner, repo_name, revision, path,
                              base_commit_id):
        """Return a file from a repository.

        This will perform an API request to fetch the contents of a file.

        Args:
            repo_owner (unicode):
                The owner of the repository.

            repo_name (unicode):
                The name of the repository.

            path (unicode):
                The file path.

            revision (unicode):
                The revision of the file to retrieve.

            base_commit_id (unicode):
                The ID of the commit that the file was changed in. This may
                not be provided, and is dependent on the type of repository.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the file contents. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.FileNotFoundError:
                The file was not found.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = self._get_file_api_url(repo_owner=repo_owner,
                                     repo_name=repo_name,
                                     revision=revision,
                                     base_commit_id=base_commit_id,
                                     path=path)

        try:
            return self.http_get(url).data
        except HostingServiceAPIError as e:
            if e.http_code == 404:
                raise FileNotFoundError(path=path,
                                        revision=revision,
                                        base_commit_id=base_commit_id)

    def api_get_file_exists(self, repo_owner, repo_name, revision, path,
                            base_commit_id):
        """Return whether a file exists in a repository.

        This will perform an API request to fetch information on the file,
        using that to determine if the file exists.

        Args:
            repo_owner (unicode):
                The owner of the repository.

            repo_name (unicode):
                The name of the repository.

            path (unicode):
                The file path.

            revision (unicode):
                The revision of the file.

            base_commit_id (unicode, optional):
                The ID of the commit that the file was changed in. This may
                not be provided, and is dependent on the type of repository.

        Returns:
            bool:
            ``True`` if the file exists. ``False`` if it does not.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the file information. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = self._get_file_api_url(repo_owner=repo_owner,
                                     repo_name=repo_name,
                                     revision=revision,
                                     base_commit_id=base_commit_id,
                                     path=path)

        try:
            self.http_head(url)
            return True
        except HostingServiceAPIError as e:
            if e.http_code == 404:
                return False

            raise

    def api_get_branches(self, repo_owner, repo_name, per_page=100, **kwargs):
        """Return a paginator of all branches for a repository.

        Args:
            repo_owner (unicode):
                The owner of the repository.

            repo_name (unicode):
                The name of the repository.

            per_page (int, optional):
                The number of branches to return per page.

            **kwargs (dict):
                Additional keyword arguments for the request.

        Returns:
            BitbucketAPIPaginator:
            A paginator for fetching branches on the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.
        """
        url = ('%s/refs/branches'
               % self._get_repositories_api_url(repo_owner, repo_name))

        return BitbucketAPIPaginator(client=self,
                                     url=url,
                                     per_page=per_page,
                                     request_kwargs=kwargs)

    def api_get_commits(self, repo_owner, repo_name, start=None, per_page=20,
                        **kwargs):
        """Return a list of commits for a repository.

        Args:
            repo_owner (unicode):
                The owner of the repository.

            repo_name (unicode):
                The name of the repository.

            start (unicode, optional):
                The optional starting commit ID or branch for the list. This
                may be used for pagination purposes.

            per_page (int, optional):
                The number of branches to return per page.

            **kwargs (dict):
                Additional keyword arguments for the request.

        Returns:
            BitbucketAPIPaginator:
            A paginator for fetching commits on the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = ('%s/commits'
               % self._get_repositories_api_url(repo_owner, repo_name))

        if start:
            url = '%s/%s' % (url, quote(start))

        return BitbucketAPIPaginator(client=self,
                                     url=url,
                                     per_page=per_page,
                                     request_kwargs=kwargs)

    def api_get_commit(self, repo_owner, repo_name, revision, **kwargs):
        """Return a commit at a given revision/commit ID.

        Args:
            repo_owner (unicode):
                The owner of the repository.

            repo_name (unicode):
                The name of the repository.

            revision (unicode):
                The revision/ID of the commit to fetch.

        Returns:
            dict:
            Information on the commit from the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = ('%s/commit/%s'
               % (self._get_repositories_api_url(repo_owner, repo_name),
                  quote(revision)))

        return self.http_get(url, **kwargs).json

    def api_get_commit_diff(self, repo_owner, repo_name, revision):
        """Return the diff for a commit at a given revision/commit ID.

        This will normalize the diff to ensure it always ends with a trailing
        newline.

        Args:
            repo_owner (unicode):
                The owner of the repository.

            repo_name (unicode):
                The name of the repository.

            revision (unicode):
                The revision/ID of the commit to fetch.

        Returns:
            dict:
            Information on the commit from the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = ('%s/diff/%s'
               % (self._get_repositories_api_url(repo_owner, repo_name),
                  quote(revision)))

        diff = self.http_get(url).data

        if not diff.endswith(b'\n'):
            diff += b'\n'

        return diff

    def build_http_request(self, query=None, only_fields=None, **kwargs):
        """Build a request object for an HTTP request.

        Args:
            query (dict, optional):
                Query arguments for the URL.

            only_fields (list, optional):
                A list of specific fields in the payload to include. All
                other fields will be excluded.

            **kwargs (dict):
                Additional keyword arguments used to build the request.

        Returns:
            reviewboard.hostingsvcs.base.http.HostingServiceHTTPRequest:
            The resulting request object for use in the HTTP request.
        """
        if only_fields:
            if query is None:
                query = {}

            query['fields'] = ','.join(only_fields)

        return super(BitbucketClient, self).build_http_request(query=query,
                                                               **kwargs)

    def process_http_error(self, request, e):
        """Process an HTTP error, raising a result.

        This will look at the error, raising a more suitable exception
        in its place.

        Args:
            request (reviewboard.hostingsvcs.base.http.
                     HostingServiceHTTPRequest, unused):
                The request that resulted in an error.

            e (urllib2.URLError):
                The error to check.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                The credentials provided were not valid.

            reviewboard.hostingsvcs.errors.HostingServiceAPIError:
                An error occurred communicating with the API. An unparsed
                payload is available.

            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an unexpected error performing the request.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        # Perform any default checks.
        super(BitbucketClient, self).process_http_error(request, e)

        if isinstance(e, HTTPError):
            data = e.read().decode('utf-8')

            try:
                rsp = json.loads(data)
            except Exception:
                rsp = None

            message = None
            detail = None

            if rsp:
                rsp_error = rsp.get('error')

                if rsp_error:
                    message = rsp_error.get('message')
                    detail = rsp_error.get('detail')

            if e.code == 401:
                raise AuthorizationError(
                    message or gettext(
                        'Invalid Bitbucket username or password. Make sure '
                        'you are using your Bitbucket username and not e-mail '
                        'address, and are using an app password if two-factor '
                        'authentication is enabled.'))
            else:
                error_message = detail or message or data

                if 'Too many invalid password attempts' in error_message:
                    raise AuthorizationError(error_message)
                else:
                    raise HostingServiceAPIError(
                        error_message,
                        http_code=e.code,
                        rsp=rsp)
        else:
            raise HostingServiceError(e.reason)

    def _get_repositories_api_url(self, repo_owner, repo_name=None):
        """Return the API URL for working with repositories.

        Args:
            repo_owner (unicode):
                The owner of the repositories.

            repo_name (unicode, optional):
                The optional name of a repository to include in the URL.

        Returns:
            unicode:
            The URL for working with a list of repositories or the specified
            repository.
        """
        url = '%s/repositories/%s' % (self.api_url, repo_owner)

        if repo_name is not None:
            url = '%s/%s' % (url, quote(repo_name))

        return url

    def _get_file_api_url(self, repo_owner, repo_name, revision,
                          base_commit_id, path):
        """Return the API URL for working with files in a repository.

        Args:
            repo_owner (unicode):
                The owner of the repositories.

            repo_name (unicode):
                The name of a repository to include in the URL.

            revision (unicode):
                The revision of the file.

            base_commit_id (unicode):
                The ID of the commit that the file was changed in. This may
                not be provided, and is dependent on the type of repository.

            path (unicode):
                The path to the file.

        Returns:
            unicode:
            The URL for working with a file in the repository.
        """
        return ('%s/src/%s/%s'
                % (self._get_repositories_api_url(repo_owner, repo_name),
                   quote(base_commit_id or revision),
                   quote(path)))


class Bitbucket(BaseHostingService):
    """Hosting service support for Bitbucket.

    Bitbucket is a hosting service that supports Git and Mercurial
    repositories, and provides issue tracker support. It's available
    at https://bitbucket.org/.
    """

    name = 'Bitbucket'
    hosting_service_id = 'bitbucket'

    client_class = BitbucketClient
    auth_form = BitbucketAuthForm

    needs_authorization = True
    supports_repositories = True
    supports_bug_trackers = True
    supports_post_commit = True

    has_repository_hook_instructions = True

    repository_url_patterns = [
        path('hooks/<str:hooks_uuid>/close-submitted/',
             BitbucketHookViews.post_receive_hook_close_submitted,
             name='bitbucket-hooks-close-submitted'),
    ]

    supported_scmtools = ['Git', 'Mercurial']
    visible_scmtools = ['Git']

    plans = [
        ('personal', {
            'name': _('Personal'),
            'form': BitbucketPersonalForm,
            'repository_fields': {
                'Git': {
                    'path': 'git@bitbucket.org:%(hosting_account_username)s/'
                            '%(bitbucket_repo_name)s.git',
                    'mirror_path': 'https://%(hosting_account_username)s@'
                                   'bitbucket.org/'
                                   '%(hosting_account_username)s/'
                                   '%(bitbucket_repo_name)s.git',
                },
                'Mercurial': {
                    'path': 'https://%(hosting_account_username)s@'
                            'bitbucket.org/%(hosting_account_username)s/'
                            '%(bitbucket_repo_name)s',
                    'mirror_path': 'ssh://hg@bitbucket.org/'
                                   '%(hosting_account_username)s/'
                                   '%(bitbucket_repo_name)s',
                },
            },
            'bug_tracker_field': ('https://bitbucket.org/'
                                  '%(hosting_account_username)s/'
                                  '%(bitbucket_repo_name)s/issue/%%s/'),
        }),
        ('other-user', {
            'name': _('Other User'),
            'form': BitbucketOtherUserForm,
            'repository_fields': {
                'Git': {
                    'path': 'git@bitbucket.org:'
                            '%(bitbucket_other_user_username)s/'
                            '%(bitbucket_other_user_repo_name)s.git',
                    'mirror_path': 'https://%(hosting_account_username)s@'
                                   'bitbucket.org/'
                                   '%(bitbucket_other_user_username)s/'
                                   '%(bitbucket_other_user_repo_name)s.git',
                },
                'Mercurial': {
                    'path': 'https://%(hosting_account_username)s@'
                            'bitbucket.org/%(bitbucket_other_user_username)s/'
                            '%(bitbucket_other_user_repo_name)s',
                    'mirror_path': 'ssh://hg@bitbucket.org/'
                                   '%(bitbucket_other_user_username)s/'
                                   '%(bitbucket_other_user_repo_name)s',
                },
            },
            'bug_tracker_field': ('https://bitbucket.org/'
                                  '%(bitbucket_other_user_username)s/'
                                  '%(bitbucket_other_user_repo_name)s/'
                                  'issue/%%s/'),
        }),
        ('team', {
            'name': _('Team'),
            'form': BitbucketTeamForm,
            'repository_fields': {
                'Git': {
                    'path': 'git@bitbucket.org:%(bitbucket_team_name)s/'
                            '%(bitbucket_team_repo_name)s.git',
                    'mirror_path': 'https://%(hosting_account_username)s@'
                                   'bitbucket.org/%(bitbucket_team_name)s/'
                                   '%(bitbucket_team_repo_name)s.git',
                },
                'Mercurial': {
                    'path': 'https://%(hosting_account_username)s@'
                            'bitbucket.org/%(bitbucket_team_name)s/'
                            '%(bitbucket_team_repo_name)s',
                    'mirror_path': 'ssh://hg@bitbucket.org/'
                                   '%(bitbucket_team_name)s/'
                                   '%(bitbucket_team_repo_name)s',
                },
            },
            'bug_tracker_field': ('https://bitbucket.org/'
                                  '%(bitbucket_team_name)s/'
                                  '%(bitbucket_team_repo_name)s/issue/%%s/'),

        }),
    ]

    DEFAULT_PLAN = 'personal'

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def check_repository(
        self,
        *,
        plan: str = DEFAULT_PLAN,
        tool_name: (str | None) = None,
        **kwargs,
    ) -> None:
        """Check the validity of a repository configuration.

        This will ensure that the configuration data being provided by the
        user is correct and doesn't contain URLs or repository names with
        ``.git`` in the name.

        It will then perform an API request against Bitbucket to get
        information on the repository. This will throw an exception if
        the repository was not found, or does not match the expected
        repository type, and return cleanly if it was found.

        Version Changed:
            8.0:
            Made arguments keyword-only.

        Args:
            plan (str, optional):
                The configured repository plan.

            tool_name (str, optional):
                The name of the tool selected to communicate with the
                repository.

            **kwargs (dict, unused):
                Additional information passed by the repository form.

        Raises:
            reviewboard.hostingsvcs.errors.RepositoryError:
                The repository configuration is not valid.

            reviewboard.scmtools.errors.RepositoryNotFoundError:
                The repository was not found.
        """
        repo_name = self._get_repository_name_raw(plan, kwargs)

        if '/' in repo_name:
            raise RepositoryError(gettext(
                'Please specify just the name of the repository, not '
                'a path.'))

        if '.git' in repo_name:
            raise RepositoryError(gettext(
                'Please specify just the name of the repository without '
                '".git".'))

        try:
            rsp = self.client.api_get_repository(
                repo_owner=self._get_repository_owner_raw(plan, kwargs),
                repo_name=self._get_repository_name_raw(plan, kwargs),
                only_fields=['scm'])
        except HostingServiceAPIError as e:
            if e.http_code == 404:
                raise RepositoryNotFoundError()

            raise

        scm = rsp['scm']

        if ((scm == 'git' and tool_name != 'Git') or
            (scm == 'hg' and tool_name != 'Mercurial')):
            raise RepositoryError(
                gettext('The Bitbucket repository being configured does not '
                        'match the type of repository you have selected.'))

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def authorize(
        self,
        *,
        username: str | None,
        password: str | None,
        **kwargs,
    ) -> None:
        """Authorize an account on Bitbucket.

        This will attempt to access the user session resource using the
        provided credentials, determining if they're valid. Those credentials
        may be one of:

        1. A username (not e-mail address) and standard password
        2. A username and an app password (recommended, and required if using
           two-factor authentication)

        If successful, the password is stored in an encrypted form.

        Version Changed:
            8.0:
            Made arguments keyword-only.

        Args:
            username (str):
                The username for the account.

            password (str):
                The user's password or app password.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.
        """
        # If this fails, an exception will be raised.
        self.client.api_get_user_session(username=username,
                                         password=password)

        self.account.data['password'] = encrypt_password(password)
        self.account.save()

    def is_authorized(self):
        """Return if the account has a stored auth token.

        This will check if we have previously stored password for the
        account. It does not validate that the credentials still work.
        """
        return self.account.data.get('password') is not None

    def get_file(self, repository, path, revision, base_commit_id=None,
                 *args, **kwargs):
        """Return a file from a repository.

        This will perform an API request to fetch the contents of a file.

        If using Git, this will expect a base commit ID to be provided.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve the file from.

            path (unicode):
                The file path.

            revision (unicode):
                The revision the file should be retrieved from.

            base_commit_id (unicode):
                The ID of the commit that the file was changed in. This may
                not be provided, and is dependent on the type of repository.

            *args (tuple, unused):
                Additional positional arguments.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.

            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be found.
        """
        if repository.tool.name == 'Git' and not base_commit_id:
            raise FileNotFoundError(
                path,
                revision,
                detail=gettext('The necessary revision information needed '
                               'to find this file was not provided. Use '
                               'RBTools 0.5.2 or newer.'))

        return self.client.api_get_file_contents(
            repo_owner=self._get_repository_owner(repository),
            repo_name=self._get_repository_name(repository),
            revision=revision,
            path=path,
            base_commit_id=base_commit_id)

    def get_file_exists(self, repository, path, revision, base_commit_id,
                        *args, **kwargs):
        """Return whether a file exists in a repository.

        This will perform an API request to fetch information on the file,
        using that to determine if the file exists.

        If using Git, this will expect a base commit ID to be provided.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve the file from.

            path (unicode):
                The file path.

            revision (unicode):
                The revision the file should be retrieved from.

            base_commit_id (unicode):
                The ID of the commit that the file was changed in. This may
                not be provided, and is dependent on the type of repository.

            *args (tuple, unused):
                Additional positional arguments.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            bool:
            ``True`` if the file exists. ``False`` if it does not.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.
        """
        if repository.tool.name == 'Git' and not base_commit_id:
            return False

        return self.client.api_get_file_exists(
            repo_owner=self._get_repository_owner(repository),
            repo_name=self._get_repository_name(repository),
            revision=revision,
            path=path,
            base_commit_id=base_commit_id)

    def get_repository_hook_instructions(self, request, repository):
        """Return instructions for setting up incoming webhooks.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            repository (reviewboard.scmtools.models.Repository):
                The repository to configure webhooks for.

        Returns:
            django.utils.safestring.SafeText:
            The HTML to display with instructions.
        """
        webhook_endpoint_url = build_server_url(local_site_reverse(
            'bitbucket-hooks-close-submitted',
            local_site=repository.local_site,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': repository.hosting_account.service_name,
                'hooks_uuid': repository.get_or_create_hooks_uuid(),
            }))
        add_webhook_url = (
            'https://bitbucket.org/%s/%s/admin/addon/admin/'
            'bitbucket-webhooks/bb-webhooks-repo-admin'
            % (self._get_repository_owner(repository),
               self._get_repository_name(repository)))

        example_id = 123
        example_url = build_server_url(local_site_reverse(
            'review-request-detail',
            local_site=repository.local_site,
            kwargs={
                'review_request_id': example_id,
            }))

        return render_to_string(
            template_name='hostingsvcs/bitbucket/repo_hook_instructions.html',
            request=request,
            context={
                'example_id': example_id,
                'example_url': example_url,
                'repository': repository,
                'server_url': get_server_url(),
                'add_webhook_url': add_webhook_url,
                'webhook_endpoint_url': webhook_endpoint_url,
            })

    def get_branches(self, repository):
        """Return all upstream branches in the repository.

        This will paginate through all the results, 100 entries at a time,
        returning all branches listed in the repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve branches from.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The list of branches found in the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.
        """
        repo_owner = self._get_repository_owner(repository)
        repo_name = self._get_repository_name(repository)

        repo_info = self.client.api_get_repository(
            repo_owner=repo_owner,
            repo_name=repo_name,
            only_fields=['mainbranch.name'])

        try:
            default_branch_name = repo_info['mainbranch']['name']
        except KeyError:
            # No default branch was set in this repository. It may be an
            # empty repository.
            return None

        branches = []
        found_default_branch = False

        paginator = self.client.api_get_branches(
            repo_owner=repo_owner,
            repo_name=repo_name,
            only_fields=['values.name', 'values.target.hash', 'next'])

        for page in paginator:
            for branch_info in page:
                try:
                    branch_name = branch_info['name']
                    is_default = (branch_name == default_branch_name)

                    if is_default:
                        found_default_branch = True

                    branches.append(Branch(
                        id=branch_name,
                        commit=branch_info['target']['hash'],
                        default=is_default))
                except KeyError as e:
                    logger.error('Missing "%s" key in Bitbucket branch '
                                 'definition %r for repository %s. Skipping '
                                 'branch.',
                                 e, branch_info, repository.pk)

        if not found_default_branch:
            branches[0].default = True

        return branches

    def get_commits(self, repository, branch=None, start=None):
        """Return a page of commits in the repository.

        This will return 20 commits at a time. The list of commits can start
        on a given branch (for branch filtering) or commit (for pagination).

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve branches from.

            branch (unicode, optional):
                The branch to retrieve commits from.

            start (unicode, optional):
                The first commit to retrieve in the page, for pagination.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The list of commits found in the repository.
        """
        paginator = self.client.api_get_commits(
            repo_owner=self._get_repository_owner(repository),
            repo_name=self._get_repository_name(repository),
            start=start or branch,
            only_fields=self._get_commit_fields(prefix='values.'))

        # Note that we're only building commits for one page worth of data.
        return [
            self._build_commit_from_rsp(commit_info)
            for commit_info in paginator.page_data
        ]

    def get_change(self, repository, revision):
        """Return a commit at a given revision/commit ID.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to fetch the commit from.

            revision (unicode):
                The revision/ID of the commit to fetch.

        Returns:
            reviewboard.scmtools.core.Commit:
            The commit from the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.
        """
        repo_owner = self._get_repository_owner(repository)
        repo_name = self._get_repository_name(repository)

        # We try to pull the commit's metadata out of the cache. The diff API
        # endpoint is just the raw content of the diff and contains no
        # metadata.
        commit = cache.get(repository.get_commit_cache_key(revision))

        if not commit:
            # However, if it is not in the cache, we have to hit the API to
            # get the metadata.
            commit_rsp = self.client.api_get_commit(
                repo_owner=repo_owner,
                repo_name=repo_name,
                revision=revision,
                only_fields=self._get_commit_fields())
            commit = self._build_commit_from_rsp(commit_rsp)

        # Now fetch the diff.
        diff = self.client.api_get_commit_diff(repo_owner=repo_owner,
                                               repo_name=repo_name,
                                               revision=revision)

        return Commit(author_name=commit.author_name,
                      id=commit.id,
                      date=commit.date,
                      message=commit.message,
                      diff=diff,
                      parent=commit.parent)

    def _get_commit_fields(self, prefix=''):
        """Return the fields needed in a query string for commit parsing.

        This is needed by APIs that want to limit the fields in the payload
        and need to parse commits.

        Args:
            prefix (unicode, optional):
                An optional prefix for each field.

        Returns:
            list of unicode:
            The fields to include in an ``only_fields=`` parameter.
        """
        return [
            prefix + name
            for name in ('author.raw', 'hash', 'date', 'message',
                         'parents.hash')
        ]

    def _build_commit_from_rsp(self, commit_rsp):
        """Return a Commit from an API reesponse.

        This will parse a response from the API and return a structured
        commit.

        Args:
            commit_rsp (dict):
                The API payload for a commit.

        Returns:
            reviewboard.scmtools.core.Commit:
            A commit based on the payload.
        """
        commit = Commit(
            author_name=commit_rsp['author']['raw'],
            id=commit_rsp['hash'],
            date=commit_rsp['date'],
            message=commit_rsp['message'])

        if commit_rsp['parents']:
            commit.parent = commit_rsp['parents'][0]['hash']

        return commit

    def _get_repository_plan(self, repository):
        """Return the stored plan for a repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository containing the stored data.

        Returns:
            unicode:
            The plan ID.
        """
        return (repository.extra_data.get('repository_plan') or
                self.DEFAULT_PLAN)

    def _get_repository_name(self, repository):
        """Return the stored Bitbucket name for a repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository containing the stored data.

        Returns:
            unicode:
            The name of the repository on Bitbucket.
        """
        return self._get_repository_name_raw(
            self._get_repository_plan(repository),
            repository.extra_data)

    def _get_repository_name_raw(self, plan, extra_data):
        """Return the Bitbucket name for a plan and repository data.

        Args:
            plan (unicode):
                The plan ID.

            extra_data (dict):
                The stored data on the repository.

        Returns:
            unicode:
            The name of the repository on Bitbucket.

        Raises:
            reviewboard.hostingsvcs.errors.InvalidPlanError:
                The provided ``plan`` value is invalid.
        """
        if plan == 'personal':
            return extra_data['bitbucket_repo_name']
        elif plan == 'team':
            return extra_data['bitbucket_team_repo_name']
        elif plan == 'other-user':
            return extra_data['bitbucket_other_user_repo_name']
        else:
            raise InvalidPlanError(plan)

    def _get_repository_owner(self, repository):
        """Return the stored owner for a repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository containing the stored data.

        Returns:
            unicode:
            The name of the repository on Bitbucket.
        """
        return self._get_repository_owner_raw(
            self._get_repository_plan(repository),
            repository.extra_data)

    def _get_repository_owner_raw(self, plan, extra_data):
        """Return the repository owner for a plan and repository data.

        Args:
            plan (unicode):
                The plan ID.

            extra_data (dict):
                The stored data on the repository.

        Returns:
            unicode:
            The owner of the repository on Bitbucket.

        Raises:
            reviewboard.hostingsvcs.errors.InvalidPlanError:
                The provided ``plan`` value is invalid.
        """
        if plan == 'personal':
            return self.account.username
        elif plan == 'team':
            return extra_data['bitbucket_team_name']
        elif plan == 'other-user':
            return extra_data['bitbucket_other_user_username']
        else:
            raise InvalidPlanError(plan)



"""Unit tests for reviewboard.licensing.provider.BaseLicenseProvider.

Version Added:
    8.0
"""

from __future__ import annotations

from datetime import datetime, timezone as tz
from uuid import uuid4

import kgb
from django.utils import timezone

from reviewboard.licensing.errors import LicenseActionError
from reviewboard.testing import TestCase
from reviewboard.licensing.license_checks import \
    ProcessCheckLicenseResultStatus
from reviewboard.licensing.tests.providers import (
    BasicTestsLicenseProvider,
    ExpiredTestsLicenseProvider,
)


class BaseLicenseProviderTests(kgb.SpyAgency, TestCase):
    """Unit tests for BaseLicenseProvider.

    Version Added:
        8.0
    """

    ######################
    # Instance variables #
    ######################

    #: A static timestamp for "now", to ease unit testing.
    now: datetime

    @classmethod
    def setUpClass(cls) -> None:
        """Set up state for all unit tests in the class.

        This will store a static timestamp for "now" and a list of license
        providers to register for the tests.
        """
        super().setUpClass()

        cls.now = datetime(2025, 4, 21, 0, 0, 0, tzinfo=tz.utc)

    @classmethod
    def tearDownClass(cls) -> None:
        """Tear down state for the unit tests."""
        super().tearDownClass()

        cls.now = None  # type: ignore

    def setUp(self) -> None:
        """Set up state for a unit test.

        This will register all the license providers and force the current
        timestamp for "now".
        """
        super().setUp()

        self.spy_on(timezone.now, op=kgb.SpyOpReturn(self.now))

    def test_call_action_with_invalid_action(self) -> None:
        """Testing BaseLicenseProvider.call_action with invalid action"""
        license_provider = ExpiredTestsLicenseProvider()

        license_info = license_provider.get_license_by_id('license1')
        assert license_info is not None

        message = 'Unsupported license action "xxx".'

        with self.assertRaisesMessage(LicenseActionError, message):
            license_provider.call_action(action_id='xxx',
                                         license_info=license_info)

    def test_action_license_update_check(self) -> None:
        """Testing BaseLicenseProvider action "license-update-check"
        """
        license_provider = BasicTestsLicenseProvider()

        license_info = license_provider.get_license_by_id('license1')
        assert license_info is not None

        result = license_provider.call_action(
            'license-update-check',
            license_info=license_info)

        self.assertEqual(result, {
            'canCheck': True,
            'checkStatusURL': 'https://example.com/license1/check/',
            'credentials': None,
            'data': {
                'license_id': 'license1',
                'something': 'special',
                'version': '1.0',
            },
            'headers': None,
            'sessionToken': 'abc123',
        })

    def test_action_license_update_check_not_supported(self) -> None:
        """Testing BaseLicenseProvider action "license-update-check" not
        supported
        """
        license_provider = ExpiredTestsLicenseProvider()

        license_info = license_provider.get_license_by_id('license1')
        assert license_info is not None

        result = license_provider.call_action(
            'license-update-check',
            license_info=license_info)

        self.assertEqual(result, {
            'canCheck': False,
        })

    def test_action_process_license_update_and_applied(self) -> None:
        """Testing BaseLicenseProvider action "process-license-update" and
        new license data applied
        """
        license_provider = BasicTestsLicenseProvider()

        license_info = license_provider.get_license_by_id('license2')
        assert license_info is not None

        result = license_provider.call_action(
            'process-license-update',
            action_data={
                'check_request_data': {
                    'license_id': 'license1',
                    'something': 'special',
                    'version': '1.0',
                },
                'check_response_data': {
                    'updated': True,
                    'version': '1.0',
                },
            },
            license_info=license_info)

        self.assertEqual(result, {
            'status': ProcessCheckLicenseResultStatus.APPLIED,
            'license_infos': {
                'license1': {
                    'actionTarget': 'basic-tests-provider:license2',
                    'actions': [
                        {
                            'actionID': 'test',
                            'label': 'Test',
                            'url': 'https://example.com/license2/',
                        },
                    ],
                    'autoRenew': False,
                    'canUploadLicense': False,
                    'expiresDate': datetime(2026, 4, 21, 0, 0, 0,
                                            tzinfo=tz.utc),
                    'expiresSoon': False,
                    'gracePeriodDaysRemaining': 0,
                    'hardExpiresDate': datetime(2026, 4, 21, 0, 0, 0,
                                                tzinfo=tz.utc),
                    'isTrial': False,
                    'licenseID': 'license2',
                    'licensedTo': 'Test User',
                    'lineItems': [],
                    'manageURL': None,
                    'noticeHTML': '',
                    'planID': 'smpbpe1',
                    'planName': 'Super Mega Power Bundle Pro Enterprise',
                    'productName': 'Test Product',
                    'status': 'licensed',
                    'summary': (
                        'License for Test Product Super Mega Power Bundle '
                        'Pro Enterprise is active'
                    ),
                },
            },
        })

    def test_action_process_license_update_and_has_latest(self) -> None:
        """Testing BaseLicenseProvider action "process-license-update" and
        already has latest data
        """
        license_provider = BasicTestsLicenseProvider()

        license_info = license_provider.get_license_by_id('license2')
        assert license_info is not None

        result = license_provider.call_action(
            'process-license-update',
            action_data={
                'action_target': 'basic-tests-provider:license2',
                'check_request_data': {
                    'license_id': 'license1',
                    'something': 'special',
                    'version': '1.0',
                },
                'check_response_data': {
                    'latest': True,
                    'version': '1.0',
                },
            },
            license_info=license_info)

        self.assertEqual(result, {
            'status': ProcessCheckLicenseResultStatus.HAS_LATEST,
            'license_infos': {
                'license1': {
                    'actionTarget': 'basic-tests-provider:license2',
                    'actions': [
                        {
                            'actionID': 'test',
                            'label': 'Test',
                            'url': 'https://example.com/license2/',
                        },
                    ],
                    'autoRenew': False,
                    'canUploadLicense': False,
                    'expiresDate': datetime(2025, 4, 26, 0, 0, 0,
                                            tzinfo=tz.utc),
                    'expiresSoon': True,
                    'gracePeriodDaysRemaining': 0,
                    'hardExpiresDate': datetime(2025, 4, 26, 0, 0, 0,
                                                tzinfo=tz.utc),
                    'isTrial': False,
                    'licenseID': 'license2',
                    'licensedTo': 'Test User',
                    'lineItems': [],
                    'manageURL': None,
                    'noticeHTML': '',
                    'planID': 'plan1',
                    'planName': 'Plan 1',
                    'productName': 'Test Product',
                    'status': 'licensed',
                    'summary': (
                        'License for Test Product Plan 1 expires in 5 days'
                    ),
                },
            },
        })

    def test_action_process_license_update_and_error(self) -> None:
        """Testing BaseLicenseProvider action "process-license-update" and
        error
        """
        license_provider = BasicTestsLicenseProvider()

        license_info = license_provider.get_license_by_id('license2')
        assert license_info is not None

        message = (
            'Error processing license update: Invalid version from '
            'the licensing server! Oh no!'
        )

        with self.assertRaisesMessage(LicenseActionError, message):
            license_provider.call_action(
                'process-license-update',
                action_data={
                    'check_request_data': {
                        'license_id': 'license1',
                        'something': 'special',
                        'version': '1.0',
                    },
                    'check_response_data': {
                        'latest': True,
                        'version': '2.0',
                    },
                },
                license_info=license_info)

    def test_action_process_license_update_no_check_data(self) -> None:
        """Testing BaseLicenseProvider action "process-license-update" and
        missing check_request_data
        """
        trace_id = '00000000-0000-0000-0000-000000000001'
        self.spy_on(uuid4, op=kgb.SpyOpReturn(trace_id))

        license_provider = BasicTestsLicenseProvider()

        license_info = license_provider.get_license_by_id('license2')
        assert license_info is not None

        message = (
            f'Missing check_request_data value for license check. This '
            f'may be an internal error or an issue with the licensing '
            f'server. Check the Review Board server logs for more '
            f'information (error ID {trace_id}).'
        )

        with self.assertRaisesMessage(LicenseActionError, message):
            license_provider.call_action(
                'process-license-update',
                license_info=license_info)

    def test_action_process_license_update_no_response_data(self) -> None:
        """Testing BaseLicenseProvider action "process-license-update" and
        missing check_response_data
        """
        trace_id = '00000000-0000-0000-0000-000000000001'
        self.spy_on(uuid4, op=kgb.SpyOpReturn(trace_id))

        license_provider = BasicTestsLicenseProvider()

        license_info = license_provider.get_license_by_id('license2')
        assert license_info is not None

        message = (
            f'Missing check_response_data value for license check. This '
            f'may be an internal error or an issue with the licensing '
            f'server. Check the Review Board server logs for more '
            f'information (error ID {trace_id}).'
        )

        with self.assertRaisesMessage(LicenseActionError, message):
            license_provider.call_action(
                'process-license-update',
                action_data={
                    'check_request_data': '"abc123"',
                },
                license_info=license_info)

    def test_action_custom(self) -> None:
        """Testing BaseLicenseProvider custom actions"""
        license_provider = BasicTestsLicenseProvider()

        license_info = license_provider.get_license_by_id('license1')
        assert license_info is not None

        result = license_provider.call_action(
            'test',
            action_data={
                'data1': 'value1',
                'data2': 123,
            },
            license_info=license_info)

        self.assertEqual(
            result,
            {
                'ok': 'yep',
                'got_action_data': {
                    'data1': 'value1',
                    'data2': 123,
                },
            })



"""SAML SSO backend.

Version Added:
    5.0
"""

from importlib import import_module

from django.urls import path, reverse
from django.utils.translation import gettext_lazy as _, gettext
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.decorators import cached_property

from reviewboard.accounts.sso.backends.base import BaseSSOBackend
from reviewboard.accounts.sso.backends.saml.forms import SAMLSettingsForm
from reviewboard.accounts.sso.backends.saml.settings import (
    DEFAULT_ATTR_EMAIL,
    DEFAULT_ATTR_FIRSTNAME,
    DEFAULT_ATTR_FULLNAME,
    DEFAULT_ATTR_LASTNAME,
    SAMLBinding,
    SAMLDigestAlgorithm,
    SAMLNameIDFormat,
    SAMLSignatureAlgorithm,)
from reviewboard.accounts.sso.backends.saml.views import (
    SAMLACSView,
    SAMLLinkUserView,
    SAMLLoginView,
    SAMLMetadataView,
    SAMLSLSView)


class SAMLSSOBackend(BaseSSOBackend):
    """SAML SSO backend.

    Version Added:
        5.0
    """

    backend_id = 'saml'
    name = _('SAML 2.0')
    settings_form = SAMLSettingsForm
    siteconfig_defaults = {
        'saml_attr_email': DEFAULT_ATTR_EMAIL,
        'saml_attr_firstname': DEFAULT_ATTR_FIRSTNAME,
        'saml_attr_fullname': DEFAULT_ATTR_FULLNAME,
        'saml_attr_lastname': DEFAULT_ATTR_LASTNAME,
        'saml_automatically_provision_users': True,
        'saml_digest_algorithm': SAMLDigestAlgorithm.SHA1,
        'saml_enabled': False,
        'saml_issuer': '',
        'saml_login_button_text': _('Log in with SAML SSO'),
        'saml_nameid_format': SAMLNameIDFormat.PERSISTENT,
        'saml_require_login_to_link': True,
        'saml_signature_algorithm': SAMLSignatureAlgorithm.DSA_SHA1,
        'saml_slo_binding_type': SAMLBinding.HTTP_REDIRECT,
        'saml_slo_url': '',
        'saml_sso_binding_type': SAMLBinding.HTTP_POST,
        'saml_sso_url': '',
        'saml_verification_cert': '',
    }
    login_view_cls = SAMLLoginView

    def __init__(self):
        """Initialize the SSO backend."""
        super().__init__()

        self._available = None

    @cached_property
    def login_label(self):
        """The text to show on the login button.

        Type:
            str
        """
        siteconfig = SiteConfiguration.objects.get_current()
        return siteconfig.get('saml_login_button_text')

    @cached_property
    def login_url(self):
        """The URL to the login page.

        Type:
            str
        """
        return reverse(
            'sso:%s:login' % self.backend_id,
            kwargs={'backend_id': self.backend_id})

    @cached_property
    def urls(self):
        """A list of URLs to register for this backend.

        Type:
            list
        """
        return [
            path('metadata/',
                 SAMLMetadataView.as_view(sso_backend=self),
                 name='metadata'),
            path('acs/',
                 SAMLACSView.as_view(sso_backend=self),
                 name='acs'),
            path('sls/',
                 SAMLSLSView.as_view(sso_backend=self),
                 name='sls'),
            path('link-user/',
                 SAMLLinkUserView.as_view(sso_backend=self),
                 name='link-user'),
        ]

    def is_available(self):
        """Return whether this backend is available.

        Returns:
            tuple:
            A two-tuple. The items in the tuple are:

            1. A bool indicating whether the backend is available.
            2. If the first element is ``False``, this will be a user-visible
               string indicating the reason why the backend is not available.
               If the backend is available, this element will be ``None``.
        """
        if self._available is None:
            try:
                import_module('onelogin.saml2.constants')
                self._available = True
            except ImportError:
                self._available = False

        if self._available:
            return True, None
        else:
            return False, gettext(
                'To enable SAML 2.0 support, install the '
                '<code>ReviewBoard[saml]</code> Python package and restart '
                'the web server.')



"""Unit tests for reviewboard.datagrids.columns.PeopleColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.contrib.auth.models import User
from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import PeopleColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class PeopleColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.PeopleColumn.

    Version Added:
        8.0
    """

    column = PeopleColumn()

    def test_render_data_with_no_people(self) -> None:
        """Testing PeopleColumn.render_data with no people"""
        review_request = self.create_review_request(publish=True)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_render_data_with_one_person(self) -> None:
        """Testing PeopleColumn.render_data with one person"""
        review_request = self.create_review_request(publish=True)
        user = User.objects.get(username='grumpy')
        review_request.target_people.add(user)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, 'grumpy')

    def test_render_data_with_multiple_people(self) -> None:
        """Testing PeopleColumn.render_data with multiple people"""
        review_request = self.create_review_request(publish=True)
        grumpy = User.objects.get(username='grumpy')
        admin = User.objects.get(username='admin')
        review_request.target_people.add(grumpy, admin)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, 'admin grumpy')

    def test_to_json_with_no_people(self) -> None:
        """Testing PeopleColumn.to_json with no people"""
        review_request = self.create_review_request(publish=True)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            [])

    def test_to_json_with_people(self) -> None:
        """Testing PeopleColumn.to_json with people"""
        review_request = self.create_review_request(publish=True)
        grumpy = User.objects.get(username='grumpy')
        admin = User.objects.get(username='admin')
        review_request.target_people.add(grumpy, admin)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            [admin, grumpy])



"""Add DiffSetHistory.last_diff_updated.

Version Added:
    1.7
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('DiffSetHistory', 'last_diff_updated', models.DateTimeField,
             null=True)
]



"""Tests for reviewboard.diffviewer.models.diffcommit."""

from __future__ import annotations

from typing import Sequence, Tuple

from django.utils import timezone

from reviewboard.diffviewer.models import DiffCommit, DiffSet
from reviewboard.testing import TestCase


class DiffCommitTests(TestCase):
    """Tests for reviewboard.diffviewer.models.diffcommit.DiffCommit."""

    fixtures = ['test_scmtools']

    def test_get_line_counts_raw(self):
        """Testing DiffCommit.get_line_counts() returns correct raw counts"""
        diffset, commits = self._populate_commits()

        self.assertEqual(diffset.get_total_line_counts(), {
            'raw_insert_count': 2,
            'raw_delete_count': 2,
            'insert_count': 2,
            'delete_count': 2,
            'replace_count': None,
            'equal_count': None,
            'total_line_count': None,
        })

        self.assertEqual(commits[0].get_total_line_counts(), {
            'raw_insert_count': 1,
            'raw_delete_count': 1,
            'insert_count': 1,
            'delete_count': 1,
            'replace_count': None,
            'equal_count': None,
            'total_line_count': None,
        })
        self.assertEqual(commits[1].get_total_line_counts(), {
            'raw_insert_count': 1,
            'raw_delete_count': 1,
            'insert_count': 1,
            'delete_count': 1,
            'replace_count': None,
            'equal_count': None,
            'total_line_count': None,
        })

    def test_get_line_counts_computed(self):
        """Testing DiffCommit.get_total_line_counts() returns correct computed
        line counts
        """
        diffset, commits = self._populate_commits()

        for filediff in diffset.files.all():
            filediff.extra_data.update({
                'insert_count': 0,
                'delete_count': 0,
                'equal_count': 0,
                'replace_count': 1,
                'total_line_count': 2,
            })
            filediff.save(update_fields=('extra_data',))

        self.assertEqual(diffset.get_total_line_counts(), {
            'raw_insert_count': 2,
            'raw_delete_count': 2,
            'insert_count': 0,
            'delete_count': 0,
            'replace_count': 2,
            'equal_count': 0,
            'total_line_count': 4,
        })

        self.assertEqual(commits[0].get_total_line_counts(), {
            'raw_insert_count': 1,
            'raw_delete_count': 1,
            'insert_count': 0,
            'delete_count': 0,
            'replace_count': 1,
            'equal_count': 0,
            'total_line_count': 2,
        })

        self.assertEqual(commits[1].get_total_line_counts(), {
            'raw_insert_count': 1,
            'raw_delete_count': 1,
            'insert_count': 0,
            'delete_count': 0,
            'replace_count': 1,
            'equal_count': 0,
            'total_line_count': 2,
        })

    def test_ordering(self):
        """Testing DiffCommits are returned in the correct order"""
        commits = self._populate_commits()[1]

        self.assertEqual(list(DiffCommit.objects.all()),
                         commits)

    def test_commit_message_body(self) -> None:
        """Testing DiffCommit.commit_message_body"""
        diffcommit = DiffCommit(commit_message=(
            'This is a summary line.\n'
            '\n'
            'This is a multi-line\n'
            'description.\n'
        ))

        self.assertEqual(
            diffcommit.commit_message_body,
            (
                'This is a multi-line\n'
                'description.'
            ))

    def test_commit_message_body_with_excess_newlines(self) -> None:
        """Testing DiffCommit.commit_message_body with newlines before and
        after body
        """
        diffcommit = DiffCommit(commit_message=(
            'This is a summary line.\n'
            '\n'
            '\n'
            'This is a multi-line\n'
            'description.\n'
            '\n'
        ))

        self.assertEqual(
            diffcommit.commit_message_body,
            (
                'This is a multi-line\n'
                'description.'
            ))

    def test_commit_message_body_with_only_summary(self) -> None:
        """Testing DiffCommit.commit_message_body with only summary"""
        diffcommit = DiffCommit(commit_message=(
            'This is a summary line.\n'
        ))

        self.assertIsNone(diffcommit.commit_message_body)

    def _populate_commits(self) -> Tuple[DiffSet, Sequence[DiffCommit]]:
        """Populate and return commits used for testing.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (reviewboard.diffviewer.models.diffset.DiffSet):
                    The parent diffset for the commits.

                1 (list):
                    The list of commits.
        """
        repository = self.create_repository()
        diffset = DiffSet.objects.create_empty(repository=repository)

        now = timezone.now()

        commits = [
            DiffCommit.objects.create(
                diffset=diffset,
                filename='diff',
                author_name='Author Name',
                author_email='author@example.com',
                commit_message='Message',
                author_date=now,
                commit_id='a' * 40,
                parent_id='0' * 40),
            DiffCommit.objects.create(
                diffset=diffset,
                filename='diff',
                author_name='Author Name',
                author_email='author@example.com',
                commit_message='Message',
                author_date=now,
                commit_id='b' * 40,
                parent_id='a' * 40),
        ]

        diff_data = self.DEFAULT_GIT_FILEDIFF_DATA_DIFF

        self.create_filediff(diffset,
                             commit=commits[0],
                             diff=diff_data)
        self.create_filediff(diffset,
                             commit=commits[1],
                             diff=diff_data)

        return diffset, commits



import inspect

from djblets.db.query import LocalDataQuerySet
from djblets.util.decorators import augment_method_from
from djblets.webapi.fields import (BooleanFieldType,
                                   ListFieldType,
                                   StringFieldType)

from reviewboard.hostingsvcs.base import (BaseHostingService,
                                          hosting_service_registry)
from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.resources import resources


class HostingServiceResource(WebAPIResource):
    """Provides information on registered hosting services.

    Review Board has a list of supported remote hosting services for
    repositories and bug trackers. These hosting services contain the
    information needed for Review Board to link with any repositories hosted
    on that service and access content for display in the diff viewer.

    This resource allows for querying that list and determining what
    capabilities of the hosting service can be used by Review Board.
    """
    added_in = '2.5'

    name = 'hosting_service'
    model_object_key = 'hosting_service_id'
    model = BaseHostingService
    uri_object_key = 'hosting_service_id'
    uri_object_key_regex = r'[a-z0-9_-]+'

    fields = {
        'id': {
            'type': StringFieldType,
            'description': "The hosting service's unique ID.",
        },
        'name': {
            'type': StringFieldType,
            'description': 'The name of the hosting service.',
        },
        'needs_authorization': {
            'type': BooleanFieldType,
            'description': 'Whether an account must be authorized and linked '
                           'in order to use this service.',
        },
        'self_hosted': {
            'type': BooleanFieldType,
            'description': 'Whether the service is meant to be self-hosted '
                           'in the network.',
        },
        'plans': {
            'type': dict,
            'description': (
                'Information on account configuration plans supported by '
                'the hosting service. These correspond to the '
                '``repository_plan`` field used when creating or updating a '
                'repository (see :ref:`webapi2.0-repository-list-resource`). '
                'This is not used for all services.'
            ),
            'added_in': '3.0.19',
        },
        'supported_scmtools': {
            'type': ListFieldType,
            'items': {
                'type': StringFieldType,
            },
            'description': 'The comprehensive list of repository types '
                           'supported by Review Board. Each of these is a '
                           'registered SCMTool ID or human-readable name.\n'
                           '\n'
                           'Some of these may not be supported by the service '
                           'anymore. See ``visible_scmtools``.'
        },
        'supports_bug_trackers': {
            'type': BooleanFieldType,
            'description': 'Whether bug trackers are available.',
        },
        'supports_list_remote_repositories': {
            'type': BooleanFieldType,
            'description': 'Whether remote repositories on the hosting '
                           'service can be listed through the API.',
        },
        'supports_repositories': {
            'type': BooleanFieldType,
            'description': 'Whether repository linking is supported.',
        },
        'supports_two_factor_auth': {
            'type': BooleanFieldType,
            'description': 'Whether two-factor authentication is supported '
                           'when linking an account.',
        },
        'visible_scmtools': {
            'type': ListFieldType,
            'items': {
                'type': StringFieldType,
            },
            'description': 'The list of repository types that are shown by '
                           'Review Board when configuring a new repository. '
                           'Each of these is a registered SCMTool ID or '
                           'human-readable name.',
            'added_in': '3.0.17',
        },
    }

    def serialize_id_field(self, hosting_service, *args, **kwargs):
        return hosting_service.hosting_service_id

    def serialize_plans_field(self, hosting_service, *args, **kwargs):
        """Serialize the plans field.

        This will convert the existing :py:attr:`BaseHostingService.plans
        <reviewboard.hostingsvcs.base.hosting_service.BaseHostingService.
        plans>` field (or create a new one if the service doesn't support
        multiple plans) into a more slimmed-down payload that can be
        transmitted via the API.

        Args:
            hosting_service (reviewboard.hostingsvcs.base.hosting_service.
                             BaseHostingService):
                The hosting service being serialized.

            *args (tuple, unused):
                Additional positional arguments.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            dict:
            The serialized plan information.
        """
        plans = hosting_service.plans
        default_form = hosting_service.form

        if not plans:
            plans = [
                ('', {
                    'name': 'Default',
                    'fields': default_form,
                }),
            ]

        return {
            plan_id: {
                'name': info['name'],
                'fields': {
                    field_name: {
                        'name': field.label,
                        'required': field.required,
                        'help_text': field.help_text,
                    }
                    for field_name, field in
                        info.get('form', default_form).base_fields.items()
                },
            }
            for plan_id, info in plans
        }

    def serialize_visible_scmtools_field(self, hosting_service, *args,
                                         **kwargs):
        """Serialize the visible_scmtools field on the hosting service.

        Args:
            hosting_service (reviewboard.hostingsvcs.base.hosting_service.
                             BaseHostingService):
                The hosting service being serialized.

            *args (tuple, unused):
                Additional positional arguments.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            list of unicode:
            The list of visible SCMTools.
        """
        # If the hosting service does not explicitly define this, it will be
        # None. We need to then return the list of supported SCMTools.
        scmtools = hosting_service.visible_scmtools

        if scmtools is None:
            scmtools = hosting_service.supported_scmtools

        return scmtools

    def has_list_access_permissions(self, *args, **kwargs):
        return True

    def has_access_permissions(self, *args, **kwargs):
        return True

    def get_queryset(self, request, *args, **kwargs):
        return LocalDataQuerySet(list(hosting_service_registry))

    def get_serializer_for_object(self, obj):
        if inspect.isclass(obj) and issubclass(obj, BaseHostingService):
            return self

        return super(HostingServiceResource, self).get_serializer_for_object(
            obj)

    def get_links(self, items, obj=None, *args, **kwargs):
        links = super(HostingServiceResource, self).get_links(
            items, obj, *args, **kwargs)

        if obj:
            request = kwargs.get('request')

            accounts_url = resources.hosting_service_account.get_list_url(
                local_site_name=request._local_site_name)
            repos_url = resources.repository.get_list_url(
                local_site_name=request._local_site_name)

            links.update({
                'accounts': {
                    'method': 'GET',
                    'href': request.build_absolute_uri(
                        '%s?service=%s' % (accounts_url,
                                           obj.hosting_service_id)
                    ),
                },
                'repositories': {
                    'method': 'GET',
                    'href': request.build_absolute_uri(
                        '%s?hosting-service=%s'
                        % (repos_url, obj.hosting_service_id)
                    ),
                }
            })

        return links

    @augment_method_from(WebAPIResource)
    def get_list(self, request, *args, **kwargs):
        """Lists all the hosting services supported by Review Board.

        Unlike most resources, this will not be paginated. The number of
        hosting services will be small, and it's useful to get them all in
        one request.
        """
        pass

    @augment_method_from(WebAPIResource)
    def get(self, request, *args, **kwargs):
        """Returns information on a particular hosting service.

        This will cover the capabilities of the hosting service, and
        information needed to help link repositories and bug trackers
        hosted on it.
        """
        pass


hosting_service_resource = HostingServiceResource()



"""SCMTool implementation for Git."""

from __future__ import annotations

import io
import logging
import os
import platform
import re
import stat
from typing import Any, Dict, TYPE_CHECKING
from urllib.parse import (quote as urlquote,
                          urlparse,
                          urlsplit as urlsplit,
                          urlunsplit as urlunsplit,
                          uses_netloc)

from django.utils.encoding import force_bytes
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _, gettext
from djblets.util.filesystem import is_exe_in_path

from reviewboard import get_manual_url
from reviewboard.diffviewer.parser import (DiffParser, DiffParserError,
                                           ParsedDiffFile)
from reviewboard.scmtools.core import SCMClient, SCMTool, HEAD, PRE_CREATION
from reviewboard.scmtools.errors import (FileNotFoundError,
                                         InvalidRevisionFormatError,
                                         RepositoryNotFoundError,
                                         SCMError)
from reviewboard.scmtools.forms import StandardSCMToolRepositoryForm
from reviewboard.ssh import utils as sshutils

if TYPE_CHECKING:
    from reviewboard.scmtools.models import Repository
    from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


GIT_DIFF_EMPTY_CHANGESET_SIZE = 3


# Register these URI schemes so we can handle them properly.
uses_netloc.append('git')


sshutils.register_rbssh('GIT_SSH')


class GitRepositoryForm(StandardSCMToolRepositoryForm):
    """Form for editing git repositories.

    Version Added:
        6.0
    """

    def clean(self) -> Dict[str, Any]:
        """Perform validation on the form.

        Returns:
            dict:
            The cleaned form data.
        """
        cleaned_data = super().clean()
        assert cleaned_data is not None

        path = cleaned_data.get('path', '')

        if path:
            url_parts = urlparse(cleaned_data['path'])

            if (url_parts.scheme not in ('', 'file') and
                not cleaned_data.get('raw_file_url', '')):
                docs_url = ('%sadmin/configuration/repositories/git/' %
                            get_manual_url())

                self.add_error('path', mark_safe(
                    gettext(
                        'Remote Git repositories cannot be accessed without a '
                        'Raw File URL Mask. See the <a href="%s">'
                        'documentation</a> for more details.')
                    % docs_url))

        return cleaned_data


class ShortSHA1Error(InvalidRevisionFormatError):
    def __init__(self, path, revision, *args, **kwargs):
        InvalidRevisionFormatError.__init__(
            self,
            path=path,
            revision=revision,
            detail=str(_('The SHA1 is too short. Make sure the diff is '
                         'generated with `git diff --full-index`.')),
            *args, **kwargs)


class GitTool(SCMTool):
    """
    You can only use this tool with a locally available git repository.
    The repository path should be to the .git directory (important if
    you do not have a bare repository).
    """

    scmtool_id = 'git'
    name = 'Git'
    diffs_use_absolute_paths = True
    supports_history = True
    commits_have_committer = True
    supports_raw_file_urls = True
    field_help_text = {
        'path': _('For local Git repositories, this should be the path to a '
                  '.git directory that Review Board can read from. For remote '
                  'Git repositories, it should be the clone URL.'),
    }
    repository_form = GitRepositoryForm
    dependencies = {
        'executables': ['git']
    }

    def __init__(
        self,
        repository: Repository,
    ) -> None:
        """Initialize the Git tool.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository owning the instance of this tool.
        """
        super().__init__(repository)

        local_site_name = None

        if repository.local_site:
            local_site_name = repository.local_site.name

        credentials = repository.get_credentials()

        self.client = GitClient(
            path=repository.path,
            raw_file_url=repository.raw_file_url,
            username=credentials['username'],
            password=credentials['password'],
            encoding=repository.encoding,
            local_site_name=local_site_name,
            local_site=repository.local_site,
        )

    def get_file(self, path, revision=HEAD, **kwargs):
        if revision == PRE_CREATION:
            return b''

        return self.client.get_file(path, revision)

    def file_exists(self, path, revision=HEAD, **kwargs):
        if revision == PRE_CREATION:
            return False

        try:
            return self.client.get_file_exists(path, revision)
        except (FileNotFoundError, InvalidRevisionFormatError):
            return False

    def normalize_patch(self, patch, filename, revision):
        """Normalize the provided patch file.

        This will update modes on new, changed, and deleted symlinks, stripping
        the symlink mode and making them appear as normal files. This will
        avoid any issues with applying the diff, and allow us to instead parse
        the symlink change as a regular file.

        Args:
            patch (bytes):
                The diff/patch file to normalize.

            filename (unicode):
                The name of the file being changed in the diff.

            revision (unicode):
                The revision of the file being changed in the diff.

        Returns:
            bytes:
            The resulting diff/patch file.
        """
        return strip_git_symlink_mode(patch)

    def parse_diff_revision(self, filename, revision, moved=False,
                            copied=False, *args, **kwargs):
        """Parse and return a filename and revision from a diff.

        Args:
            filename (bytes):
                The filename as represented in the diff.

            revision (bytes):
                The revision as represented in the diff.

            moved (bool, optional):
                Whether this is a moved file.

            copied (bool, optional):
                Whether this is a copied file.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            tuple:
            A tuple containing two items:

            1. The normalized filename as a byte string.
            2. The normalized revision as a byte string or a
               :py:class:`~reviewboard.scmtools.core.Revision`.

        Raises:
            ShortSHA1Error:
                The revision is a short SHA-1.
        """
        assert isinstance(filename, bytes), (
            'filename must be a byte string, not %s' % type(filename))
        assert isinstance(revision, bytes), (
            'revision must be a byte string, not %s' % type(revision))

        if filename == b'/dev/null':
            revision = PRE_CREATION
        elif (revision != PRE_CREATION and
              (not (moved or copied) or revision != b'')):
            # Moved files with no changes have no revision, so don't validate
            # those.
            self.client.validate_sha1_format(filename, revision)

        return filename, revision

    def get_parser(self, data):
        return GitDiffParser(data)

    @classmethod
    def check_repository(
        cls,
        path: str,
        username: (str | None) = None,
        password: (str | None) = None,
        local_site_name: (str | None) = None,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> None:
        """Check a repository configuration for validity.

        This should check if a repository exists and can be connected to.
        This will also check if the repository requires an HTTPS certificate.

        A failed result is returned as an exception. The exception may contain
        extra information, such as a human-readable description of the problem.
        If the repository is valid and can be connected to, no exception will
        be thrown.

        Args:
            path (str):
                The repository path.

            username (str, optional):
                The optional username for the repository.

            password (str, optional):
                The optional password for the repository.

            local_site_name (str, optional):
                The name of the :term:`Local Site` that owns this repository.
                This is optional.

            local_site (reviewboard.site.models.LocalSite, optional):
                The :term:`Local Site` that owns this repository.

            **kwargs (dict, unused):
                Additional settings for the repository.

        Raises:
            reviewboard.scmtools.errors.AuthenticationError:
                The provided username/password or the configured SSH key could
                not be used to authenticate with the repository.

            reviewboard.scmtools.errors.RepositoryNotFoundError:
                A repository could not be found at the given path.

            reviewboard.scmtools.errors.SCMError:
                There was a generic error with the repository or its
                configuration.  Details will be provided in the error message.

            reviewboard.ssh.errors.BadHostKeyError:
                An SSH path was provided, but the host key for the repository
                did not match the expected key.

            reviewboard.ssh.errors.SSHError:
                An SSH path was provided, but there was an error establishing
                the SSH connection.

            reviewboard.ssh.errors.SSHInvalidPortError:
                An SSH path was provided, but the port specified was not a
                valid number.

            Exception:
                An unexpected exception has occurred. Callers should check
                for this and handle it.
        """
        client = GitClient(
            path=path,
            username=username,
            password=password,
            local_site_name=local_site_name,
            local_site=local_site,
        )

        super().check_repository(
            path=client.path,
            username=username,
            password=password,
            local_site_name=local_site_name,
            local_site=local_site,
            **kwargs,
        )

        if not client.is_valid_repository():
            raise RepositoryNotFoundError()

        # TODO: Check for an HTTPS certificate. This will require pycurl.


class GitDiffParser(DiffParser):
    """Diff parser for standard Git diffs.

    Version Changed:
        4.0.6:
        Added storage of UNIX file modes and symlink targets.
    """

    pre_creation_regexp = re.compile(b"^0+$")

    FILE_MODE_RE = re.compile(
        br'^(?:(?P<type>new|old|deleted) (?:file )?mode|index \w+\.\.\w+) '
        br'(?P<mode>\d+)$',
        re.M)

    DIFF_GIT_LINE_RES = [
        # Match with a/ and b/ prefixes. Common case.
        re.compile(
            br'^diff --git'
            br' (?P<aq>")?a/(?P<orig_filename>[^"]+)(?(aq)")'
            br' (?P<bq>")?b/(?P<new_filename>[^"]+)(?(bq)")$'),

        # Match without a/ and b/ prefixes. Spaces are allowed only if using
        # quotes around the filename.
        re.compile(
            br'^diff --git'
            br' (?P<aq>")?(?!a/)(?P<orig_filename>(?(aq)[^"]|[^ ])+)(?(aq)")'
            br' (?P<bq>")?(?!b/)(?P<new_filename>(?(bq)[^"]|[^ ])+)(?(bq)")$'),

        # Match without a/ and b/ prefixes, without quotes, and with the
        # original and new names being identical.
        re.compile(
            br'^diff --git'
            br' (?!")(?!a/)(?P<orig_filename>[^"]+)(?!")'
            br' (?!")(?!b/)(?P<new_filename>(?P=orig_filename))(?!")$'),
    ]

    EXTENDED_HEADERS_KEYS = set([
        b'old mode',
        b'new mode',
        b'deleted file mode',
        b'new file mode',
        b'copy from',
        b'copy to',
        b'rename from',
        b'rename to',
        b'similarity index',
        b'dissimilarity index',
        b'index',
    ])

    def _parse_extended_headers(self, linenum):
        """Parse an extended headers section.

        A dictionary with keys being the header name and values
        being a tuple of (header value, complete header line) will
        be returned. The complete header lines will have a trailing
        new line added for convenience.
        """
        headers = {}

        while linenum < len(self.lines):
            line = self.lines[linenum]

            for key in self.EXTENDED_HEADERS_KEYS:
                if line.startswith(key):
                    headers[key] = line[len(key) + 1:], line + b'\n'
                    break
            else:
                # No headers were found on this line so we're
                # done parsing them.
                break

            linenum += 1

        return headers, linenum

    def parse(self):
        """
        Parses the diff, returning a list of File objects representing each
        file in the diff.
        """
        self.files = []
        i = 0
        preamble = io.BytesIO()

        while i < len(self.lines):
            next_i, file_info, new_diff = self._parse_diff(i)

            if file_info:
                if self.files:
                    self.files[-1].append_data(preamble.getvalue())
                    preamble.close()
                    preamble = io.BytesIO()
                    self.files[-1].finalize()

                self._ensure_file_has_required_fields(file_info)

                file_info.prepend_data(preamble.getvalue())
                preamble.close()
                preamble = io.BytesIO()

                self.files.append(file_info)
            elif new_diff:
                # We found a diff, but it was empty and has no file entry.
                # Reset the preamble.
                preamble.close()
                preamble = io.BytesIO()
            else:
                preamble.write(self.lines[i])
                preamble.write(b'\n')

            i = next_i

        try:
            if self.files:
                self.files[-1].append_data(preamble.getvalue())
                self.files[-1].finalize()
            elif preamble.getvalue().strip() != b'':
                # This is probably not an actual git diff file.
                raise DiffParserError('This does not appear to be a git diff',
                                      0)
        finally:
            preamble.close()

        return self.files

    def _parse_diff(self, linenum):
        """Parses out one file from a Git diff

        This will return a tuple of the next line number, the file info
        (if any), and whether or not we've found a file (even if we decided
        not to record it).
        """
        if self.lines[linenum].startswith(b"diff --git"):
            line, info = self._parse_git_diff(linenum)

            return line, info, True
        else:
            return linenum + 1, None, False

    def _parse_git_diff(self, linenum):
        """Parse a Git-style diff header.

        This will parse a diff header containing file mode information,
        file operations, and ``diff --git`` lines, and filename information.

        Args:
            linenum (int):
                The current line number.

        Returns:
            tuple:
            A tuple containing the following:

            1. The next line number to parse.
            2. The populated :py:class:`ParsedDiffFile` instance for this
               file, if any.
        """
        lines = self.lines
        diff_git_line = lines[linenum]
        linenum += 1

        # Check to make sure we haven't reached the end of the diff.
        if linenum >= len(lines):
            return linenum, None

        file_info = ParsedDiffFile(parsed_diff_change=self.parsed_diff_change)
        file_info.append_data(diff_git_line)
        file_info.append_data(b'\n')
        file_info.binary = False

        # Assume the blob / commit information is provided globally. If
        # we found an index header we'll override this.
        file_info.orig_file_details = self.base_commit_id
        file_info.modified_file_details = self.new_commit_id

        headers, linenum = self._parse_extended_headers(linenum)

        # Determine the created/deleted/modified state and accompanying UNIX
        # file mode.
        if self._is_new_file(headers):
            new_mode_header = headers[b'new file mode'][1]
            file_info.append_data(new_mode_header)
            file_info.orig_file_details = PRE_CREATION
            file_info.new_unix_mode = self._parse_unix_mode(new_mode_header)
        elif self._is_deleted_file(headers):
            old_mode_header = headers[b'deleted file mode'][1]
            file_info.append_data(old_mode_header)
            file_info.deleted = True
            file_info.old_unix_mode = self._parse_unix_mode(old_mode_header)
        elif self._is_mode_change(headers):
            old_mode_header = headers[b'old mode'][1]
            new_mode_header = headers[b'new mode'][1]
            file_info.append_data(old_mode_header)
            file_info.append_data(new_mode_header)
            file_info.old_unix_mode = self._parse_unix_mode(old_mode_header)
            file_info.new_unix_mode = self._parse_unix_mode(new_mode_header)

        # Determine whether the file has been moved or copied, and track
        # that information.
        if self._is_moved_file(headers):
            file_info.orig_filename = headers[b'rename from'][0]
            file_info.modified_filename = headers[b'rename to'][0]
            file_info.moved = True

            if b'similarity index' in headers:
                file_info.append_data(headers[b'similarity index'][1])

            file_info.append_data(headers[b'rename from'][1])
            file_info.append_data(headers[b'rename to'][1])
        elif self._is_copied_file(headers):
            file_info.orig_filename = headers[b'copy from'][0]
            file_info.modified_filename = headers[b'copy to'][0]
            file_info.copied = True

            if b'similarity index' in headers:
                file_info.append_data(headers[b'similarity index'][1])

            file_info.append_data(headers[b'copy from'][1])
            file_info.append_data(headers[b'copy to'][1])

        # Assume by default that the change is empty. If we find content
        # later, we'll clear this.
        empty_change = True

        if b'index' in headers:
            index_header_pair = headers[b'index']
            index_range = index_header_pair[0].split()[0]
            index_header = index_header_pair[1]

            if b'..' in index_range:
                (file_info.orig_file_details,
                 file_info.modified_file_details) = index_range.split(b'..')

            if self.pre_creation_regexp.match(file_info.orig_file_details):
                file_info.orig_file_details = PRE_CREATION

            file_info.append_data(index_header)
            unix_mode = self._parse_unix_mode(index_header)

            if unix_mode is not None:
                # This will overwrite anything set above. In theory, a Git
                # diff shouldn't have multiple (conflicting) mode lines.
                file_info.old_unix_mode = unix_mode
                file_info.new_unix_mode = unix_mode

        changes_linenum = None

        # Get the changes
        while linenum < len(lines):
            if self._is_git_diff(linenum):
                break
            elif self._is_binary_patch(linenum):
                file_info.binary = True
                file_info.append_data(lines[linenum])
                file_info.append_data(b'\n')
                empty_change = False
                linenum += 1
                break
            elif self._is_diff_fromfile_line(linenum):
                orig_line = lines[linenum]
                new_line = lines[linenum + 1]

                orig_filename = orig_line[len(b'--- '):]
                new_filename = new_line[len(b'+++ '):]

                # Some diffs may incorrectly contain filenames listed as:
                #
                # --- filename\t
                # +++ filename\t
                #
                # We need to strip those single trailing tabs.
                if orig_filename.endswith(b'\t'):
                    orig_filename = orig_filename[:-1]

                if new_filename.endswith(b'\t'):
                    new_filename = new_filename[:-1]

                # Strip the Git a/ and b/ prefixes, if set in the diff.
                if orig_filename.startswith(b'a/'):
                    orig_filename = orig_filename[2:]

                if new_filename.startswith(b'b/'):
                    new_filename = new_filename[2:]

                if orig_filename == b'/dev/null':
                    file_info.orig_file_details = PRE_CREATION
                    file_info.orig_filename = new_filename
                else:
                    file_info.orig_filename = orig_filename

                if new_filename == b'/dev/null':
                    file_info.modified_filename = orig_filename
                else:
                    file_info.modified_filename = new_filename

                file_info.append_data(orig_line)
                file_info.append_data(b'\n')
                file_info.append_data(new_line)
                file_info.append_data(b'\n')

                linenum += 2
                changes_linenum = linenum
            else:
                empty_change = False
                linenum = self.parse_diff_line(linenum, file_info)

        # Now that we have the UNIX file mode and changed lines, we can
        # determine if this is a symlink. We need to check the new and old
        # UNIX modes.
        mode_for_symlink = (file_info.new_unix_mode or
                            file_info.old_unix_mode)

        if (mode_for_symlink is not None and
            stat.S_ISLNK(int(mode_for_symlink, 8))):
            file_info.is_symlink = True

            if changes_linenum is not None:
                for i in range(changes_linenum, linenum):
                    line = lines[i]

                    if line.startswith(b'-'):
                        file_info.old_symlink_target = line[1:].strip()
                    elif line.startswith(b'+'):
                        file_info.new_symlink_target = line[1:].strip()

        if not file_info.orig_filename:
            # This file didn't have any --- or +++ lines. This usually means
            # the file was deleted or moved without changes. We'll need to
            # fall back to parsing the diff --git line, which is more
            # error-prone.
            assert not file_info.modified_filename

            self._parse_diff_git_line(diff_git_line, file_info, linenum)

        # For an empty change, we keep the file's info only if it is a new
        # 0-length file, a moved file, a copied file, or a deleted 0-length
        # file.
        #
        # TODO: In the future, we'll want to keep empty files so we can show
        #       metadata changes, once that functionality is available in the
        #       diff viewer.
        if (empty_change and
            file_info.orig_file_details != PRE_CREATION and
            not (file_info.moved or file_info.copied or file_info.deleted)):
            # We didn't find any interesting content, so leave out this
            # file's info.
            #
            # Note that we may want to change this in the future to preserve
            # data like mode changes, but that will require filtering out
            # empty changes at the diff viewer level in a sane way.
            file_info.discard()
            file_info = None

        return linenum, file_info

    def _parse_diff_git_line(self, diff_git_line, file_info, linenum):
        """Parses the "diff --git" line for filename information.

        Not all diffs have "---" and "+++" lines we can parse for the
        filenames. Git leaves these out if there aren't any changes made
        to the file.

        This function attempts to extract this information from the
        "diff --git" lines in the diff. It supports the following:

        * All filenames with quotes.
        * All filenames with a/ and b/ prefixes.
        * Filenames without quotes, prefixes, or spaces.
        * Filenames without quotes or prefixes, where the original and
          modified filenames are identical.
        """
        for regex in self.DIFF_GIT_LINE_RES:
            m = regex.match(diff_git_line)

            if m:
                file_info.orig_filename = m.group('orig_filename')
                file_info.modified_filename = m.group('new_filename')
                return

        raise DiffParserError(
            'Unable to parse the "diff --git" line for this file, due to '
            'the use of filenames with spaces or --no-prefix, --src-prefix, '
            'or --dst-prefix options.',
            linenum)

    def _parse_unix_mode(self, line):
        """Return a UNIX file mode from a line, if found.

        This will parse the provided line, looking for a UNIX file mode.

        Args:
            line (bytes):
                The line to parse.

        Returns:
            bytes:
            The file mode, or ``None`` if not found.
        """
        m = self.FILE_MODE_RE.match(line)

        if m:
            return m.group('mode').decode('utf-8')

        return None

    def _is_new_file(self, headers):
        return b'new file mode' in headers

    def _is_deleted_file(self, headers):
        return b'deleted file mode' in headers

    def _is_mode_change(self, headers):
        return b'old mode' in headers and b'new mode' in headers

    def _is_copied_file(self, headers):
        return b'copy from' in headers and b'copy to' in headers

    def _is_moved_file(self, headers):
        return b'rename from' in headers and b'rename to' in headers

    def _is_git_diff(self, linenum):
        return self.lines[linenum].startswith(b'diff --git')

    def _is_binary_patch(self, linenum):
        line = self.lines[linenum]

        return (line.startswith(b"Binary file") or
                line.startswith(b"GIT binary patch"))

    def _is_diff_fromfile_line(self, linenum):
        return (linenum + 1 < len(self.lines) and
                (self.lines[linenum].startswith(b'--- ') and
                    self.lines[linenum + 1].startswith(b'+++ ')))

    def _ensure_file_has_required_fields(self, file_info):
        """Make sure that the file object has all expected fields.

        This is needed so that there aren't explosions higher up the chain when
        the web layer is expecting a string object.
        """
        for attr in ('orig_file_details', 'modified_file_details'):
            if getattr(file_info, attr) is None:
                setattr(file_info, attr, b'')


class GitClient(SCMClient):
    FULL_SHA1_LENGTH = 40

    schemeless_url_re = re.compile(
        r'^(?P<username>[A-Za-z0-9_\.-]+@)?(?P<hostname>[A-Za-z0-9_\.-]+):'
        r'(?P<path>.*)')

    def __init__(
        self,
        path: str,
        raw_file_url: (str | None) = None,
        username: (str | None) = None,
        password: (str | None) = None,
        encoding: str = '',
        local_site_name: (str | None) = None,
        **kwargs,
    ) -> None:
        """Initialize the Git client.

        Args:
            path (str):
                The configured repository path.

            raw_file_url (str, optional):
                The optional URL template for accessing raw files via HTTP(S).

            username (str, optional):
                The optional username for communicating with the repository.

            password (str, optional):
                The optional password for communicating with the repository.

            encoding (str, optional):
                An optional encoding.

                This is unused and scheduled for deprecation.

            local_site_name (str, optional):
                The name of the associated Local Site, if any.

            **kwargs (dict):
                Keyword arguments to pass to the parent method.

        Raises:
            ImportError:
                There was an error locating the :command:`git` command line
                tool.

            reviewboard.scmtools.errors.SCMError:
                There was an error accessing the repository.
        """
        super().__init__(
            self._normalize_git_url(path),
            username=username,
            password=password,
            **kwargs,
        )

        if not is_exe_in_path('git'):
            # This is technically not the right kind of error, but it's the
            # pattern we use with all the other tools.
            raise ImportError

        self.raw_file_url = raw_file_url
        self.encoding = encoding
        self.local_site_name = local_site_name
        self.git_dir = None

        url_parts = urlparse(self.path)

        if url_parts[0] == 'file':
            if platform.system() == "Windows":
                # Windows requires drive letter (e.g. C:/)
                self.git_dir = url_parts[1] + url_parts[2]
            else:
                self.git_dir = url_parts[2]

            with self._run_git(
                ['--git-dir=%s' % self.git_dir, 'config',
                 'core.repositoryformatversion'],
            ) as p:
                failure = p.wait()

            if failure:
                # See if we have a permissions error
                if not os.access(self.git_dir, os.R_OK):
                    raise SCMError(_("Permission denied accessing the local "
                                     "Git repository '%s'") % self.git_dir)
                else:
                    raise SCMError(_('Unable to retrieve information from '
                                     'local Git repository'))

    def is_valid_repository(self):
        """Checks if this is a valid Git repository."""
        url_parts = urlsplit(self.path)

        if (url_parts.scheme.lower() in ('http', 'https') and
            url_parts.username is None and self.username):
            # Git URLs, especially HTTP(s), that require authentication should
            # be entered without the authentication info in the URL (because
            # then it would be visible), but we need it in the URL when testing
            # to make sure it exists. Reformat the path here to include them.
            new_netloc = urlquote(self.username, safe='')

            if self.password:
                new_netloc += ':' + urlquote(self.password, safe='')

            new_netloc += '@' + url_parts.netloc

            path = urlunsplit((url_parts[0], new_netloc, url_parts[2],
                               url_parts[3], url_parts[4]))
        else:
            path = self.path

        with self._run_git(['ls-remote', path, 'HEAD']) as p:
            errmsg = p.stderr.read()
            failure = p.wait()

        if failure:
            logger.error('Git: Failed to find valid repository %s: %s',
                         self.path, errmsg)
            return False

        return True

    def get_file(self, path, revision):
        if self.raw_file_url:
            self.validate_sha1_format(path, revision)

            return self.get_file_http(self._build_raw_url(path, revision),
                                      path, revision)
        else:
            return self._cat_file(path, revision, "blob")

    def get_file_exists(self, path, revision):
        if self.raw_file_url:
            try:
                # We want to make sure we can access the file successfully,
                # without any HTTP errors. A successful access means the file
                # exists. The contents themselves are meaningless, so ignore
                # them. If we do successfully get the file without triggering
                # any sort of exception, then the file exists.
                self.get_file(path, revision)
                return True
            except Exception:
                return False
        else:
            contents = self._cat_file(path, revision, '-t')
            return contents and contents.strip() == b'blob'

    def validate_sha1_format(self, path, sha1):
        """Validates that a SHA1 is of the right length for this repository."""
        if self.raw_file_url and len(sha1) != self.FULL_SHA1_LENGTH:
            raise ShortSHA1Error(path, sha1)

    def _run_git(self, args):
        """Runs a git command, returning a subprocess.Popen."""
        return SCMTool.popen(['git'] + args,
                             local_site_name=self.local_site_name)

    def _build_raw_url(self, path, revision):
        url = self.raw_file_url
        url = url.replace("<revision>", revision)
        url = url.replace("<filename>", urlquote(path))
        return url

    def _cat_file(self, path, revision, option):
        """
        Call git-cat-file(1) to get content or type information for a
        repository object.

        If called with just "commit", gets the content of a blob (or
        raises an exception if the commit is not a blob).

        Otherwise, "option" can be used to pass a switch to git-cat-file,
        e.g. to test or existence or get the type of "commit".
        """
        commit = self._resolve_head(revision, path)

        with self._run_git(['--git-dir=%s' % self.git_dir, 'cat-file',
                            option, commit]) as p:
            contents = force_bytes(p.stdout.read())
            errmsg = force_bytes(p.stderr.read())
            failure = p.wait()

        if failure:
            if errmsg.startswith(b'fatal: Not a valid object name'):
                raise FileNotFoundError(path, revision=commit)
            else:
                raise SCMError(errmsg.decode('utf-8'))

        return contents

    def _resolve_head(self, revision, path):
        if revision == HEAD:
            if path == "":
                raise SCMError("path must be supplied if revision is %s"
                               % HEAD)
            return "HEAD:%s" % path
        else:
            return str(revision)

    def _normalize_git_url(self, path):
        if path.startswith('file://'):
            return path

        url_parts = urlparse(path)
        scheme = url_parts[0]
        netloc = url_parts[1]

        if scheme and netloc:
            return path

        m = self.schemeless_url_re.match(path)

        if m:
            path = m.group('path')

            if not path.startswith('/'):
                path = '/' + path

            return 'ssh://%s%s%s' % (m.group('username') or '',
                                     m.group('hostname'),
                                     path)

        return "file://" + path


def strip_git_symlink_mode(patch_data):
    """Strip symlink modes from a file in a patch.

    This will update modes on new, changed, and deleted symlinks, stripping
    the symlink mode and making them appear as normal files. This will
    avoid any issues with applying the diff, and allow us to instead parse
    the symlink change as a regular file.

    Version Added:
        4.0.6

    Args:
        patch_data (bytes):
            The patch data.

    Returns:
        bytes:
        The normalized patch data.
    """
    m = GitDiffParser.FILE_MODE_RE.search(patch_data)

    if m:
        mode = int(m.group('mode'), 8)

        if stat.S_ISLNK(mode):
            mode = stat.S_IFREG | stat.S_IMODE(mode)
            patch_data = b'%s%o%s' % (patch_data[:m.start('mode')],
                                      mode,
                                      patch_data[m.end('mode'):])

    return patch_data



"""Unit tests for the SCMTools registry."""

from djblets.registries.errors import AlreadyRegisteredError, ItemLookupError
from djblets.testing.decorators import add_fixtures

from reviewboard.scmtools import scmtools_registry
from reviewboard.scmtools.core import SCMTool
from reviewboard.scmtools.models import Tool
from reviewboard.scmtools.registry import logger
from reviewboard.testing import TestCase


class DummySCMTool(SCMTool):
    name = 'Dummy'
    scmtool_id = 'dummy-scmtool'


class DummySCMToolWithLookupName(SCMTool):
    name = 'New Dummy'
    lookup_name = 'Old Dummy'
    scmtool_id = 'lookup-dummy-scmtool'


class SCMToolRegistryTests(TestCase):
    """Unit tests for the SCMTools registry."""

    def tearDown(self):
        """Tear down the test suite."""
        super().tearDown()

        for scmtool_id in ('dummy-scmtool', 'lookup-dummy-scmtool'):
            try:
                scmtools_registry.unregister_by_attr('scmtool_id', scmtool_id)
            except ItemLookupError:
                pass

    def test_register(self):
        """Testing SCMToolRegistry.register"""
        scmtools_registry.register(DummySCMTool)

        with self.assertRaises(AlreadyRegisteredError):
            scmtools_registry.register(DummySCMTool)

    def test_register_creates_tools(self):
        """Testing SCMToolRegistry.register creates Tool entries"""
        scmtools_registry.register(DummySCMTool)

        tool = Tool.objects.get(name='Dummy')
        self.assertEqual(
            tool.class_name,
            'reviewboard.scmtools.tests.test_registry.DummySCMTool')

    def test_register_with_lookup_name(self):
        """Testing SCMToolRegistry.register creates Tool entries with Legacy DB
        name
        """
        scmtools_registry.register(DummySCMToolWithLookupName)

        tool = Tool.objects.get(name='Old Dummy')
        self.assertEqual(
            tool.class_name,
            'reviewboard.scmtools.tests.test_registry.'
            'DummySCMToolWithLookupName')

        self.assertFalse(Tool.objects.filter(name='New Dummy').exists())

    def test_lookup(self):
        """Testing SCMToolRegistry look-up operations"""
        scmtools_registry.register(DummySCMTool)

        self.assertIs(scmtools_registry.get_by_id('dummy-scmtool'),
                      DummySCMTool)
        self.assertIs(scmtools_registry.get_by_name('Dummy'),
                      DummySCMTool)
        self.assertIs(
            scmtools_registry.get_by_class_name(
                'reviewboard.scmtools.tests.test_registry.DummySCMTool'),
            DummySCMTool)

    def test_lookup_with_lookup_name(self):
        """Testing SCMToolRegistry look-up operations with
        SCMTool.lookup_name
        """
        scmtools_registry.register(DummySCMToolWithLookupName)

        self.assertIs(
            scmtools_registry.get_by_id('lookup-dummy-scmtool'),
            DummySCMToolWithLookupName)
        self.assertIs(
            scmtools_registry.get_by_name('Old Dummy'),
            DummySCMToolWithLookupName)
        self.assertIsNone(scmtools_registry.get_by_name('New Dummy'))
        self.assertIs(
            scmtools_registry.get_by_class_name(
                'reviewboard.scmtools.tests.test_registry.'
                'DummySCMToolWithLookupName'),
            DummySCMToolWithLookupName)

    @add_fixtures(['test_scmtools'])
    def test_populate_db_with_conflicting_tools(self):
        """Testing SCMToolRegistry.populate_db with conflicting Tool entries"""
        git = Tool.objects.get(name='Git')
        perforce = Tool.objects.get(name='Perforce')

        git.name = 'Git!'
        git.save(update_fields=('name',))

        perforce.class_name = 'reviewboard.scmtools.XXXOtherPerforceTool'
        perforce.save(update_fields=('class_name',))

        with self.assertLogs(logger=logger) as log_ctx:
            scmtools_registry.populate_db()

        self.assertEqual(len(log_ctx.records), 2)
        self.assertEqual(
            log_ctx.records[0].getMessage(),
            "Tool ID %s (name='Git!', "
            "class_name='reviewboard.scmtools.git.GitTool') conflicts with "
            "SCMTool 'git' (lookup_name='Git', "
            "class_name='reviewboard.scmtools.git.GitTool')"
            % git.pk)
        self.assertEqual(
            log_ctx.records[1].getMessage(),
            "Tool ID %s (name='Perforce', "
            "class_name='reviewboard.scmtools.XXXOtherPerforceTool') "
            "conflicts with SCMTool 'perforce' (lookup_name='Perforce', "
            "class_name='reviewboard.scmtools.perforce.PerforceTool')"
            % perforce.pk)



"""Hosting service for JIRA."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django import forms
from django.utils.translation import gettext_lazy as _
try:
    from jira.client import JIRA as JIRAClient
    from jira.exceptions import JIRAError
    has_jira = True
except ImportError:
    has_jira = False

from reviewboard.admin.validation import validate_bug_tracker_base_hosting_url
from reviewboard.hostingsvcs.base.bug_tracker import BaseBugTracker
from reviewboard.hostingsvcs.base.forms import BaseHostingServiceRepositoryForm
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService

if TYPE_CHECKING:
    from reviewboard.hostingsvcs.base.bug_tracker import BugInfo
    from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


class JIRAForm(BaseHostingServiceRepositoryForm):
    jira_url = forms.CharField(
        label=_('JIRA URL'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        validators=[validate_bug_tracker_base_hosting_url])

    def clean_jira_url(self):
        return self.cleaned_data['jira_url'].rstrip('/ ')


class JIRA(BaseHostingService, BaseBugTracker):
    """Hosting service for JIRA."""

    name = 'JIRA'
    hosting_service_id = 'jira'
    form = JIRAForm
    bug_tracker_field = '%(jira_url)s/browse/%%s'
    supports_bug_trackers = True

    def __init__(self, account):
        super(JIRA, self).__init__(account)

        self.jira_client = None

    def get_bug_info_uncached(
        self,
        repository: Repository,
        bug_id: str,
    ) -> BugInfo:
        """Return the information for the specified bug.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository object.

            bug_id (str):
                The ID of the bug to fetch.

        Returns:
            reviewboard.hostingsvcs.base.bug_tracker.BugInfo:
            Information about the bug.
        """
        result: BugInfo = {
            'summary': '',
            'description': '',
            'status': '',
        }

        if has_jira:
            if not self.jira_client:
                try:
                    jira_url = repository.extra_data['bug_tracker-jira_url']
                    self.jira_client = JIRAClient(options={
                        'server': jira_url,
                    }, max_retries=0)
                except ValueError as e:
                    logger.warning(
                        'Unable to initialize JIRAClient for server %s: %s',
                        repository.extra_data['bug_tracker-jira_url'], e)
                    return result

            try:
                jira_issue = self.jira_client.issue(bug_id)
                result = {
                    'description': jira_issue.fields.description,
                    'summary': jira_issue.fields.summary,
                    'status': jira_issue.fields.status
                }
            except JIRAError as e:
                logger.warning('Unable to fetch JIRA data for issue %s: %s',
                               bug_id, e, exc_info=True)

        return result
